import argparse
import contextlib
import datetime
import errno
import json
import os
import pathlib
import tempfile
import typing
import sys

import anndata
import rich.console
import rich.progress
import numpy
import pandas
import scipy.sparse
from scipy.cluster.hierarchy import fcluster

try:
    import argcomplete
except ImportError as err:
    argcomplete = err

try:
    from rich_argparse import ArgumentDefaultsRichHelpFormatter as HelpFormatter
except ImportError:
    from argparse import ArgumentDefaultsHelpFormatter as HelpFormatter

from .. import __version__
from ..dataset.base import BaseDataset
from ..dataset.list import DatasetList
from ..mmseqs import MMSeqs, Database, Clustering
from ..pipeline import PipelineParameters, Pipeline
from ..clustering import default_strategy, ClusteringStrategy, LinearClustering, HierarchicalClustering
from .._utils import Stopwatch
from .inputs import (
    GenBankInput, 
    GenBankListInput, 
    AntiSMASHGenBankInput, 
    AntiSMASHZipInput,
    DefenseFinderTSV,
)


def build_parser(argv: typing.List[str]) -> argparse.ArgumentParser:
    show_all = '--help-all' in argv
    params = PipelineParameters.default()
    strategy = default_strategy()

    parser = argparse.ArgumentParser(
        prog="igua",
        formatter_class=HelpFormatter,
        add_help=False,
        description=(
            "A method for content-agnostic high-throughput identification of "
            "Gene Cluster Families (GCFs) from gene clusters of genomic and "
            "metagenomic origin. "
        ),
        epilog="Use --help-all to see all available options including advanced parameters." if not show_all else None,
    )

    def extended_help_text(text: str) -> str:
        return text if show_all else argparse.SUPPRESS

    def extended_parser_group(name, help) -> str:
        if show_all:
            return parser.add_argument_group(name, help)
        else:
            return parser

    parser.add_argument(
        "-h",
        "--help",
        help="Display this help message and exit.",
        action="help",
    )
    parser.add_argument(
        "--help-all",
        help="Display all options including advanced parameters.",
        action="help",
    )

    parser.add_argument(
        "-V",
        "--version",
        help="Display the program version and exit.",
        action="version",
        version=f"igua v{__version__}",
    )
    parser.add_argument(
        "-j",
        "--jobs",
        help="The number of threads to use in parallel sections.",
        type=int,
        default=os.cpu_count(),
        metavar="N",
    )

    group_input = parser.add_argument_group(
        "Input",
        "Mandatory input files required by the pipeline."
    )
    group_input.add_argument(
        "-g",
        "--gbk",
        help="Input path to GenBank files containing generic gene clusters.",
        metavar="FILE",
        dest="inputs",
        action="append",
        type=GenBankInput,
        default=[],
    )
    group_input.add_argument(
        "-l",
        "--gbk-list",
        help="Path to a file containing a list of GenBank files to process.",
        metavar="FILE",
        dest="inputs",
        action="append",
        type=GenBankListInput,
        default=[],
    )
    group_input.add_argument(
        "--antismash-gbk",
        help="Input path to GenBank files containing antiSMASH predictions.",
        metavar="FILE",
        dest="inputs",
        action="append",
        type=AntiSMASHGenBankInput,
        default=[],
    )
    group_input.add_argument(
        "--antismash-zip",
        help="Input path to Zip files containing antiSMASH predictions.",
        metavar="FILE",
        dest="inputs",
        action="append",
        type=AntiSMASHZipInput,
        default=[],
    )
    group_input.add_argument(
        "--defense-finder-tsv",
        help="Input path to TSV metadata file for DefenseFinder predictions.",
        metavar="FILE",
        dest="inputs",
        action="append",
        type=DefenseFinderTSV,
        default=[],
    )

    group_output = parser.add_argument_group(
        "Output", "Output files generated by the pipeline."
    )
    group_output.add_argument(
        "-o",
        "--output",
        help="The name of the output table to generate.",
        default=pathlib.Path("gcfs.tsv"),
        type=pathlib.Path,
    )
    group_output.add_argument(
        "-C",
        "--compositions",
        help=extended_help_text("A file where to write compositional data for GCF representatives."),
        type=pathlib.Path,
    )
    group_output.add_argument(
        "-F",
        "--features",
        help=extended_help_text("A file where to write protein cluster representatives in FASTA format."),
        type=pathlib.Path,
    )

    group_parameters = parser.add_argument_group(
        "Parameters", "General purpose parameters."
    )
    group_parameters.add_argument(
        "-w",
        "--workdir",
        help=extended_help_text("A folder to use for temporary data produced by MMSeqs2."),
    )
    group_parameters.add_argument(
        "--prefix",
        help="The prefix for GCF identifiers generated by the pipeline.",
        default="GCF",
    )

    group_mmseqs_dedup = extended_parser_group(
        "MMSeqs2 Deduplication",
        "Parameters for the first nucleotide clustering step (exact/near-exact deduplication).",
    )
    group_mmseqs_dedup.add_argument(
        "--dedup-identity",
        help=extended_help_text("Sequence identity threshold for deduplication step."),
        type=float,
        default=params.nuc1["sequence_identity"],
        metavar="FLOAT",
    )
    group_mmseqs_dedup.add_argument(
        "--dedup-coverage",
        help=extended_help_text("Coverage threshold for deduplication step."),
        type=float,
        default=params.nuc1["coverage"],
        metavar="FLOAT",
    )
    group_mmseqs_dedup.add_argument(
        "--dedup-evalue",
        help=extended_help_text("E-value threshold for deduplication step."),
        type=float,
        default=params.nuc1["e_value"],
        metavar="FLOAT",
    )
    group_mmseqs_dedup.add_argument(
        "--dedup-cluster-mode",
        help=extended_help_text("Clustering mode for deduplication: 0=SetCover, 1=Connected component, 2=Greedy incremental."),
        type=int,
        default=params.nuc1["cluster_mode"],
        choices=[0, 1, 2],
        metavar="INT",
    )
    group_mmseqs_dedup.add_argument(
        "--dedup-coverage-mode",
        help=extended_help_text("Coverage mode for deduplication: 0=target, 1=query, 2=both, 3=length of target, 4=length of query, 5=length of both."),
        type=int,
        default=params.nuc1["coverage_mode"],
        choices=[0, 1, 2, 3, 4, 5],
        metavar="INT",
    )
    group_mmseqs_dedup.add_argument(
        "--dedup-spaced-kmer-mode",
        help=extended_help_text("Spaced k-mer mode for deduplication: 0=use ungapped k-mers, 1=use spaced k-mers."),
        type=int,
        default=params.nuc1["spaced_kmer_mode"],
        choices=[0, 1],
        metavar="INT",
    )

    group_mmseqs_nuc = extended_parser_group(
        "MMSeqs2 Nucleotide Clustering",
        "Parameters for the second nucleotide clustering step (relaxed clustering of representatives).",
    )
    group_mmseqs_nuc.add_argument(
        "--nuc-identity",
        help=extended_help_text("Sequence identity threshold for nucleotide clustering step."),
        type=float,
        default=params.nuc2["sequence_identity"],
        metavar="FLOAT",
    )
    group_mmseqs_nuc.add_argument(
        "--nuc-coverage",
        help=extended_help_text("Coverage threshold for nucleotide clustering step."),
        type=float,
        default=params.nuc2["coverage"],
        metavar="FLOAT",
    )
    group_mmseqs_nuc.add_argument(
        "--nuc-evalue",
        help=extended_help_text("E-value threshold for nucleotide clustering step."),
        type=float,
        default=params.nuc2["e_value"],
        metavar="FLOAT",
    )
    group_mmseqs_nuc.add_argument(
        "--nuc-cluster-mode",
        help=extended_help_text("Clustering mode for nucleotide step: 0=SetCover, 1=Connected component, 2=Greedy incremental."),
        type=int,
        default=params.nuc2["cluster_mode"],
        choices=[0, 1, 2],
        metavar="INT",
    )
    group_mmseqs_nuc.add_argument(
        "--nuc-coverage-mode",
        help=extended_help_text("Coverage mode for nucleotide step: 0=target, 1=query, 2=both, 3=length of target, 4=length of query, 5=length of both."),
        type=int,
        default=params.nuc2["coverage_mode"],
        choices=[0, 1, 2, 3, 4, 5],
        metavar="INT",
    )
    group_mmseqs_nuc.add_argument(
        "--nuc-spaced-kmer-mode",
        help=extended_help_text("Spaced k-mer mode for nucleotide step: 0=use ungapped k-mers, 1=use spaced k-mers."),
        type=int,
        default=params.nuc2["spaced_kmer_mode"],
        choices=[0, 1],
        metavar="INT",
    )

    group_mmseqs_prot = extended_parser_group(
        "MMSeqs2 Protein Clustering",
        "Parameters for protein clustering step (used for compositional analysis).",
    )
    group_mmseqs_prot.add_argument(
        "--prot-identity",
        help=extended_help_text("Sequence identity threshold for protein clustering step."),
        type=float,
        default=params.prot["sequence_identity"],
        metavar="FLOAT",
    )
    group_mmseqs_prot.add_argument(
        "--prot-coverage",
        help=extended_help_text("Coverage threshold for protein clustering step."),
        type=float,
        default=params.prot["coverage"],
        metavar="FLOAT",
    )
    group_mmseqs_prot.add_argument(
        "--prot-evalue",
        help=extended_help_text("E-value threshold for protein clustering step."),
        type=float,
        default=params.prot["e_value"],
        metavar="FLOAT",
    )
    group_mmseqs_prot.add_argument(
        "--prot-coverage-mode",
        help=extended_help_text("Coverage mode for protein step: 0=target, 1=query, 2=both, 3=length of target, 4=length of query, 5=length of both."),
        type=int,
        default=params.prot["coverage_mode"],
        choices=[0, 1, 2, 3, 4, 5],
        metavar="INT",
    )

    group_clustering = extended_parser_group(
        "Clustering", "Parameters to control the hierarchical clustering."
    )
    group_clustering.add_argument(
        "--clustering-method",
        help=extended_help_text("The hierarchical clustering method to use for protein-level clustering."),
        default=strategy.method,
        choices=[
            "average",
            "single",
            "complete",
            "weighted",
            "centroid",
            "median",
            "ward",
            "linclust",
            "none",
        ],
    )
    group_clustering.add_argument(
        "--clustering-distance",
        help=extended_help_text("The distance threshold after which to stop merging clusters."),
        type=float,
        default=strategy.distance,
    )
    group_clustering.add_argument(
        "--clustering-precision",
        help=extended_help_text("The numerical precision to use for computing distances for hierarchical clustering."),
        default=strategy.precision,
        choices=["half", "single", "double"],
    )
    group_clustering.add_argument(
        "--clustering-weight",
        help=extended_help_text("The method to use to weigh dimensions in the distance matrix for hierarchical clustering."),
        default="protein",
        choices=["protein", "none"],
    )

    return parser

def _build_parser_sphinx():
    return build_parser(["igua", "--help-all"])


def create_dataset(
    progress: rich.progress.Progress,
    args: argparse.Namespace,
) -> BaseDataset:
    """Create dataset handler with appropriate adapter.
    """
    datasets = []
    for input_ in args.inputs:
        dataset = input_.to_dataset(args)
        datasets.append(dataset)
    return DatasetList(datasets)


def get_mmseqs_params(args: argparse.Namespace) -> PipelineParameters:
    """Build MMSeqs2 parameter dictionaries from command-line arguments."""

    params_nuc1 = dict(
        e_value=args.dedup_evalue,
        sequence_identity=args.dedup_identity,
        coverage=args.dedup_coverage,
        cluster_mode=args.dedup_cluster_mode,
        coverage_mode=args.dedup_coverage_mode,
        spaced_kmer_mode=args.dedup_spaced_kmer_mode,
    )

    params_nuc2 = dict(
        e_value=args.nuc_evalue,
        sequence_identity=args.nuc_identity,
        coverage=args.nuc_coverage,
        cluster_mode=args.nuc_cluster_mode,
        coverage_mode=args.nuc_coverage_mode,
        spaced_kmer_mode=args.nuc_spaced_kmer_mode,
    )

    params_prot = dict(
        e_value=args.prot_evalue,
        coverage=args.prot_coverage,
        coverage_mode=args.prot_coverage_mode,
        sequence_identity=args.prot_identity,
    )

    return PipelineParameters(
        nuc1=params_nuc1,
        nuc2=params_nuc2,
        prot=params_prot,
    )


def get_clustering_strategy(args: argparse.Namespace) -> ClusteringStrategy:
    if args.clustering_method == "none":
        return None
    elif args.clustering_method == "linclust":
        return LinearClustering(
            distance=args.clustering_distance,
        )
    else:
        return HierarchicalClustering(
            method=args.clustering_method,
            distance=args.clustering_distance,
            precision=args.clustering_precision,
            jobs=args.jobs,
        )


def get_protein_representative(
    mmseqs: MMSeqs,
    input_path: pathlib.Path,
    output_prefix: pathlib.Path,
    fasta_path: pathlib.Path,
) -> None:
    db = Database(mmseqs, input_path.with_suffix(".db"))
    result = Clustering(mmseqs, output_prefix.with_suffix(".db"), db)
    subdb = result.to_subdb(output_prefix.with_name(f"{output_prefix.name}_rep_seq.db"))
    subdb.to_fasta(fasta_path)


def report_completion(
    stopwatch: Stopwatch,
    console: rich.console.Console,
) -> None:
    end_time_str = stopwatch.stop_datetime.strftime("%Y-%m-%d at %H:%M:%S")
    time_str = stopwatch.total_human()

    console.print(f"[bold blue]{'Completed':>12}[/] on {end_time_str}")
    console.print(f"[bold green]{'Total time':>12}[/] {time_str}")


def main(argv: typing.Optional[typing.List[str]] = None) -> int:
    # retrieve argv
    argv = sys.argv[1:] if argv is None else argv

    # build parser and get arguments
    parser = build_parser(argv)
    if not isinstance(argcomplete, ImportError):
        argcomplete.autocomplete(parser)
    args = parser.parse_args(argv)

    with contextlib.ExitStack() as ctx:

        # use a stopwatch to measure time
        stopwatch = ctx.enter_context(Stopwatch())

        # extract MMseqs parameters from command line
        params = get_mmseqs_params(args)
        strategy = get_clustering_strategy(args)

        # use user provided workdir or create a new one in `tempfile`
        if args.workdir is None:
            tempdir = tempfile.TemporaryDirectory(prefix="igua-")
        else:
            tempdir = pathlib.Path(args.workdir)
            tempdir.mkdir(parents=True, exist_ok=True)
            tempdir = tempfile.TemporaryDirectory(prefix="igua-", dir=args.workdir)

        # validate individual file arguments
        if not args.inputs:
            parser.error("No input files given.")

        # acquire temporary folder
        workdir = pathlib.Path(ctx.enter_context(tempdir))

        # prepare progress bar
        progress = ctx.enter_context(
            rich.progress.Progress(
                "",
                rich.progress.SpinnerColumn(),
                *rich.progress.Progress.get_default_columns(),
            )
        )
        progress.console.print(f"[bold blue]{'Started':>12}[/] on {stopwatch.start_datetime.strftime('%Y-%m-%d at %H:%M:%S')}")

        # check mmseqs version (and if an mmseqs binary is available)
        mmseqs = MMSeqs(progress=progress, threads=args.jobs, tempdir=workdir)
        try:
            v = mmseqs.version()
            progress.console.print(f"[bold green]{'Using':>12}[/] [purple]MMSeqs2[/] version {v!r}")
        except RuntimeError:
            progress.console.print(f"[bold red]{'Failed':>12}[/] to locate [purple]MMSeqs2[/] binary")
            return errno.ENOENT

        # create a pipeline with the configuration from the CLI
        pipeline = Pipeline(
            strategy=strategy,
            params=params,
            mmseqs=mmseqs,
            workdir=workdir,
            progress=progress,
            weight=None if args.clustering_weight == "none" else args.clustering_weight,
        )

        # create appropriate dataset handler
        dataset = create_dataset(progress=progress, args=args)

        # run pipeline and retrieve GCFs
        result = pipeline.run(dataset)

        # save GCFs
        result.gcfs.to_csv(args.output, sep="\t", index=False)
        progress.console.print(
            f"[bold green]{'Saved':>12}[/] final GCFs table to {str(args.output)!r}"
        )

        # save compositions
        if args.compositions is not None:
            gcf_representatives = result.gcfs["gcf_representative"].unique().tolist()
            representatives_compositions = anndata.AnnData(
                X=result.compositions[gcf_representatives].X,
                var=result.compositions.var,
                obs=(
                    result.gcfs.set_index("cluster_id")
                        .loc[gcf_representatives, ["gcf_id", "gcf_representative"]]
                        .set_index("gcf_id")
                ),
            )
            representatives_compositions.write(args.compositions)
            progress.console.print(
                f"[bold green]{'Saved':>12}[/] GCF compositions to {str(args.compositions)!r}"
            )

        # save protein features
        if args.features is not None:
            get_protein_representative(
                mmseqs,
                result.proteins_faa,
                workdir.joinpath("step3"),
                args.features,
            )
            progress.console.print(
                f"[bold green]{'Saved':>12}[/] protein features to {str(args.features)!r}"
            )

    # report runtime before exiting
    report_completion(stopwatch, progress.console)

    return 0



import pandas as pd
import anndata
import pytest


test_cases = ["ecoli", "ecoli_gzip"]


@pytest.mark.parametrize("ref_data_id", test_cases)
def test_regression_gcfs(ref_data_id):
    """Compare output against known good reference"""
    # using defaults (see makefile):
    gcfs = pd.read_csv(f"tests/fixtures/{ref_data_id}_gcfs.tsv", sep="\t")
    gcfs_alternative = pd.read_csv(
        f"tests/fixtures/{ref_data_id}_gcfs_alternative.tsv", sep="\t"
    )
    current_gcfs = pd.read_csv(f"tests/test_output/{ref_data_id}_gcfs.tsv", sep="\t")

    # same shape and number of unique GCFs
    assert len(gcfs) == len(current_gcfs)
    assert gcfs["gcf_id"].nunique() == current_gcfs["gcf_id"].nunique()

    # clustering should be deterministic (if using same parameters)
    # therefore the number of unique values per column should be the same
    pd.testing.assert_series_equal(current_gcfs.nunique(), gcfs.nunique())
    # and the output gcfs should be identical
    try:
        pd.testing.assert_frame_equal(
            gcfs.sort_values("cluster_id").reset_index(drop=True),
            current_gcfs.sort_values("cluster_id").reset_index(drop=True),
        )
    except AssertionError:
        pd.testing.assert_frame_equal(
            gcfs_alternative.sort_values("cluster_id").reset_index(drop=True),
            current_gcfs.sort_values("cluster_id").reset_index(drop=True),
        )


@pytest.mark.parametrize("ref_data_id", test_cases)
def test_regression_compositions(ref_data_id):
    """Test compositions output structure"""
    comp = anndata.read_h5ad(f"tests/fixtures/{ref_data_id}_compositions.h5ad")
    comp_alternative = anndata.read_h5ad(
        f"tests/fixtures/{ref_data_id}_compositions_alternative.h5ad"
    )
    current_comp = anndata.read_h5ad(
        f"tests/test_output/{ref_data_id}_compositions.h5ad"
    )

    assert (
        comp.shape == current_comp.shape or comp_alternative.shape == current_comp.shape
    )
    assert all(comp.var_names == current_comp.var_names) or all(
        comp_alternative.var_names == current_comp.var_names
    )
    assert all(comp.obs_names == current_comp.obs_names) or all(
        comp_alternative.obs_names == current_comp.obs_names
    )



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Imports -----------------------------------------------------------------

import datetime
import os
import sys
import re
import shutil
import semantic_version
import urllib.request

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import igua

project = 'IGUA'
author = 'Martin Larralde'
year = datetime.date.today().year
copyright = '{}, {}'.format("2026" if year==2026 else "2026-{}".format(year), author)

# avoid using rich_argparse as it emits ANSI control codes that conflict
# with `sphinx-argparse` rendering of the usage
import igua.cli
import argparse
import unittest.mock
igua.cli.HelpFormatter = argparse.ArgumentDefaultsHelpFormatter

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.todo",
    "sphinx.ext.extlinks",
    "sphinxarg.ext",
    "sphinx_design",
    "sphinxcontrib.jquery",
    "nbsphinx",
    "recommonmark",
    "IPython.sphinxext.ipython_console_highlighting",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = [
    '_build',
    'Thumbs.db',
    '.DS_Store',
    '**.ipynb_checkpoints',
    'requirements.txt'
]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "monokailight"

# The name of the default role for inline references
default_role = "py:obj"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'pydata_sphinx_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static/js', '_static/bibtex', '_static/css']
html_js_files = ["custom-icon.js"]
html_css_files = ["custom.css"]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
html_theme_options = {
    "external_links": [
        {
            "url": "https://www.biorxiv.org/content/10.1101/2025.05.15.654203v1",
            "name": "Preprint",
        },
    ],
    "show_toc_level": 2,
    "use_edit_page_button": True,
    "icon_links": [
        {
            "name": "GitHub",
            "url": "https://github.com/zellerlab/IGUA",
            "icon": "fa-brands fa-github",
        },
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/igua",
            "icon": "fa-custom fa-pypi",
        },
    ],
    "logo": {
        "text": "IGUA",
        "image_light": "_images/logo.png",
        "image_dark": "_images/logo.png",
    },
    "navbar_align": "left",
    "footer_start": ["copyright"],
    "footer_center": ["sphinx-version"],
}

html_context = {
    "github_user": "zellerlab",
    "github_repo": "IGUA",
    "github_version": "main",
    "doc_path": "docs",
}

html_favicon = '_images/favicon.ico'

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "IGUA"


# -- Extension configuration -------------------------------------------------

# -- Options for extlinks extension ------------------------------------------

extlinks = {
    'doi': ('https://doi.org/%s', 'doi:%s'),
    'pmid': ('https://pubmed.ncbi.nlm.nih.gov/%s', 'PMID:%s'),
    'isbn': ('https://www.worldcat.org/isbn/%s', 'ISBN:%s'),
}

# -- Options for imgmath extension -------------------------------------------

imgmath_image_format = "svg"

# -- Options for napoleon extension ------------------------------------------

napoleon_include_init_with_doc = True
napoleon_include_special_with_doc = True
napoleon_include_private_with_doc = True
napoleon_use_admonition_for_examples = True
napoleon_use_admonition_for_notes = True
napoleon_use_admonition_for_references = True
napoleon_use_rtype = False

# -- Options for autodoc extension -------------------------------------------

autoclass_content = "class"
autodoc_member_order = 'groupwise'
autosummary_generate = []
autodoc_typehints = 'none'

# -- Options for intersphinx extension ---------------------------------------

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "psutil": ("https://psutil.readthedocs.io/en/latest/", None),
    "biopython": ("https://biopython.org/docs/latest/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "rich": ("https://rich.readthedocs.io/en/latest/", None),
    "anndata": ("https://anndata.readthedocs.io/en/latest/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
}

# -- Options for recommonmark extension --------------------------------------

source_suffix = {
    '.rst': 'restructuredtext',
    '.txt': 'markdown',
    '.md': 'markdown',
}

# -- Options for nbsphinx extension ------------------------------------------

nbsphinx_execute = 'auto'
nbsphinx_execute_arguments = [
    "--InlineBackend.figure_formats={'svg', 'pdf'}",
    "--InlineBackend.rc={'figure.dpi': 96}",
]



import pandas as pd
import anndata
from pathlib import Path
import pytest


test_cases = ["ecoli", "ecoli_gzip"]


@pytest.mark.parametrize("ref_data_id", test_cases)
def test_gcfs_output_structure(ref_data_id):
    """Test gcfs output maintains expected structure"""
    gcfs = pd.read_csv(f"tests/test_output/{ref_data_id}_gcfs.tsv", sep="\t")

    # file structure tests
    assert set(gcfs.columns) == {
        "cluster_id",
        "cluster_length",
        "gcf_id",
        "gcf_representative",
        "nucleotide_representative",
        "fragment_representative",
        "filename",
    }

    assert len(gcfs) > 0
    assert gcfs["gcf_id"].nunique() <= len(gcfs)  # gcfs <= inputs
    assert all(gcfs["cluster_length"] > 0)
    assert gcfs["gcf_id"].str.startswith("GCF").all()


@pytest.mark.parametrize("ref_data_id", test_cases)
def test_compositions_output(ref_data_id):
    """Test compositions output structure"""
    comp = anndata.read_h5ad(f"tests/test_output/{ref_data_id}_compositions.h5ad")

    assert comp.X.shape[0] > 0  # type: ignore # has gcfs
    assert comp.X.shape[1] > 0  # type: ignore # has features
    # assert "gcf_id" in comp.obs.columns
    # assert "size" in comp.var.columns


@pytest.mark.parametrize("ref_data_id", test_cases)
def test_features_output(ref_data_id):
    """Test protein features FASTA"""
    features_path = Path(f"tests/test_output/{ref_data_id}_features.fa")
    assert features_path.exists()
    assert features_path.stat().st_size > 0

    # count seqs
    with open(features_path) as f:
        seq_count = sum(1 for line in f if line.startswith(">"))
    assert seq_count > 0



import abc
import dataclasses
import pathlib
import typing

import Bio.Seq
import pandas as pd
import rich.progress

from ..mmseqs import Database, MMSeqs


@dataclasses.dataclass
class Cluster:
    """A gene cluster.

    Attributes:
        id (`str`): The identifier of the gene cluster. Must be unique 
            across all datasets, as the identifier will be used to refer
            to each cluster in the resulting gene cluster families.
        sequence (`str`): The genomic sequence of the dataset.
        source (`str` or `None`): The source of the gene cluster, 
            usually the path to the file where this cluster record
            was extracted from.

    """
    id: str
    sequence: str
    source: typing.Optional[str] = dataclasses.field(default=None)

    @property
    def length(self):
        """`int`: The length of the cluster sequence.
        """
        return len(self.sequence)


@dataclasses.dataclass
class Protein:
    """A protein inside a gene cluster.

    Attributes:
        id (`str`): The identifier of the protein. Must be unique 
            across all datasets, as the identifier will be used to refer
            to each protein while clustering.
        sequence (`str`): The amino-acid sequence of the protein.
        cluster_id (`str`): The identifier of the cluster this protein
            belongs to.

    """
    id: str
    sequence: str
    cluster_id: str

    @property
    def length(self):
        """`int`: The length of the protein sequence.
        """
        return len(self.sequence)


class BaseDataset(abc.ABC):
    """An abstract dataset to provide clusters to a `ClusteringPipeline`.

    IGUA needs the genomic sequences of the gene clusters to process, and
    the individual proteins of each gene cluster. To provide them to 
    a clustering pipeline, the `BaseDataset` interface provides two 
    functions: `BaseDataset.extract_clusters` and 
    `BaseDataset.extract_proteins` which yield `Cluster` and `Protein`
    objects, respectively.

    """

    @abc.abstractmethod
    def extract_clusters(
        self, 
        progress: rich.progress.Progress,
    ) -> typing.Iterable[Cluster]:
        """Extract the clusters from the dataset.

        Arguments:
            progress (`rich.progress.Progress`): A `Progress` instance
                that can be used for tracking progress.

        Yields:
            `Cluster`: A cluster object for each gene cluster to be 
            processed in the dataset.

        """

    @abc.abstractmethod
    def extract_proteins(
        self,
        progress: rich.progress.Progress,
        cluster_ids: typing.Collection[str],
    ) -> typing.Iterable[Protein]:
        """Extracts protein sequences from GenBank files.

        Arguments:
            progress (`rich.progress.Progress`): A `Progress` instance
                that can be used for tracking progress.
            cluster_ids (`collections.abc.Collection` of `str`): A 
                collection of cluster IDs from which to extract proteins.

        Yields:
            `Protein`: A protein object for each protein of the gene 
            clusters to be processed in the dataset.

        """




import contextlib
import datetime
import io
import pathlib
import time
from typing import BinaryIO, Iterator, Union

try:
    from isal import igzip as gzip
except ImportError:
    import gzip

try:
    import lz4.frame
except ImportError as err:
    lz4 = err

try:
    import bz2
except ImportError as err:
    bz2 = err

try:
    import lzma
except ImportError as err:
    lzma = err


class Stopwatch:
    """A stopwatch class to time execution of a context block.
    """

    def __init__(self):
        self.start_time = None
        self.start_datetime = None
        self.stop_time = None
        self.stop_datetime = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()
        return False

    def start(self):
        """Start the stopwatch.
        """
        self.start_time = time.monotonic()
        self.start_datetime = datetime.datetime.now()

    def stop(self):
        """Stop the stopwatch.
        """
        self.stop_time = time.monotonic()
        self.stop_datetime = datetime.datetime.now()

    def total(self):
        """Compute the total execution time.

        Returns: 
            `float`: The total time the stopwatch has been active
            for. May have subsecond resolution if `time.monotonic`
            supports it on the current operating system.

        """
        if self.start_time is None:
            raise RuntimeError("stopwatch has not been started")
        if self.stop_time is None:
            raise RuntimeError("stopwatch has not been stopped")
        return self.stop_time - self.start_time

    def total_human(self):
        """Format the total execution time in a human-readable format.

        Returns:
            `str`: The total execution time summarized in hours, minutes,
            seconds depending on the total runtime.
            
        """
        total_seconds = int(self.total())

        hours, remainder = divmod(total_seconds, 3600)
        minutes, seconds = divmod(remainder, 60)

        if hours > 0:
            return f"{hours}h {minutes}m {seconds}s"
        elif minutes > 0:
            return f"{minutes}m {seconds}s"
        else:
            return f"{seconds}s"


_BZ2_MAGIC = b"BZh"
_GZIP_MAGIC = b"\x1f\x8b"
_XZ_MAGIC = b"\xfd7zXZ"
_LZ4_MAGIC = b"\x04\x22\x4d\x18"

@contextlib.contextmanager
def zopen(path: Union[str, pathlib.Path, BinaryIO]) -> Iterator[BinaryIO]:
    """Open a file with optional compression in binary mode.
    """
    with contextlib.ExitStack() as ctx:
        if isinstance(path, (str, pathlib.Path)):
            file = ctx.enter_context(open(path, "rb"))
        else:
            file = ctx.enter_context(io.BufferedReader(path))
        peek = file.peek()
        if peek.startswith(_GZIP_MAGIC):
            file = ctx.enter_context(gzip.open(file, mode="rb"))
        elif peek.startswith(_BZ2_MAGIC):
            if isinstance(bz2, ImportError):
                raise RuntimeError("File compression is LZMA but lzma is not available") from lz4
            file = ctx.enter_context(bz2.open(file, mode="rb"))
        elif peek.startswith(_XZ_MAGIC):
            if isinstance(lzma, ImportError):
                raise RuntimeError("File compression is LZMA but lzma is not available") from lz4
            file = ctx.enter_context(lzma.open(file, mode="rb"))
        elif peek.startswith(_LZ4_MAGIC):
            if isinstance(lz4, ImportError):
                raise RuntimeError("File compression is LZ4 but python-lz4 is not installed") from lz4
            file = ctx.enter_context(lz4.frame.open(file))
        yield file


import datetime
import unittest

from igua._utils import Stopwatch


class TestStopwatch(unittest.TestCase):

    def test_error_not_started(self):
        stopwatch = Stopwatch()
        self.assertRaises(RuntimeError, stopwatch.total)

    def test_error_not_stopped(self):
        stopwatch = Stopwatch()
        stopwatch.start()
        self.assertRaises(RuntimeError, stopwatch.total)

    def test_total_human(self):
        stopwatch = Stopwatch()
        stopwatch.start_time = 1
        # stopwatch.start_datetime = datetime.datetime.now()
        stopwatch.stop_time = stopwatch.start_time + 2*3600 + 1
        # stopwatch.stop_datetime = datetime.datetime.now()
        self.assertEqual(stopwatch.total_human(), "2h 0m 1s")

        stopwatch.start_time = 1
        # stopwatch.start_datetime = datetime.datetime.now()
        stopwatch.stop_time = stopwatch.start_time + 2*60 + 1
        # stopwatch.stop_datetime = datetime.datetime.now()
        self.assertEqual(stopwatch.total_human(), "2m 1s")


import abc
import contextlib
import copy
import dataclasses
import pathlib
import tempfile
import typing
from typing import Dict, Literal, Mapping, Tuple

import anndata
import rich.console
import rich.progress
import scipy.sparse
import numpy
import pandas

from .dataset.base import BaseDataset
from .dataset.defensefinder import DefenseFinderDataset
from .dataset.fasta_gff import FastaGFFDataset
from .mmseqs import MMSeqs, Database, Clustering
from .clustering import ClusteringStrategy, HierarchicalClustering, LinearClustering, default_strategy
from ._utils import Stopwatch


class _BaseSink(abc.ABC):

    def __init__(self):
        self.stats = []
        self.done = set()

    def add_record(self, name: str, sequence: str, **kwargs) -> bool:
        if name not in self.done:
            self.stats.append({"id": name, "length": len(sequence), **kwargs})
            self.done.add(name)
            return True
        else:
            return False

    def report_statistic(self) -> pandas.DataFrame:
        return pandas.DataFrame(self.stats)


class _FASTASink(_BaseSink):

    def __init__(self, file: typing.TextIO) -> None:
        super().__init__()
        self.file = file

    def add_record(self, name: str, sequence: str, **kwargs) -> None:
        if not super().add_record(name, sequence, **kwargs):
            return False
        self.file.write(">")
        self.file.write(name)
        self.file.write("\n")
        self.file.write(sequence)
        self.file.write("\n")
        return True


@dataclasses.dataclass
class PipelineParameters:
    """The parameters of the IGUA pipeline.

    Attributes:
        nuc1 (`dict`): A dictionary of parameters to pass to MMseqs2 for the
            stage 1 (nucleotide deduplication).
        nuc2 (`dict`): A dictionary of parameters to pass to MMseqs2 for the
            stage 2 (nucleotide clustering).
        prot (`dict`): A dictionary of parameters to pass to MMseqs2 for the
            stage 3 (protein clustering).
        clustering_method (`str`): The name of the clustering method to
            use on the stage 3, either one of the supported hierarchical
            clustering mode, or ``linclust`` to use MMseqs2-derived linear
            clustering.
        clustering_distance (`float`): The distance cutoff to use for
            cluster creation in stage 3.
        clustering_precision (`str`): The precision to use to perform
            hierarchical clustering: one of *half*, *single*, or *double*.
        clustering_weight (`str`): The weights to use for computing pairwise
            distances during stage 3: either *protein* to weigh each
            dimension by the length of the protein it represents, or `None`
            to use uniform weights.

    """
    nuc1: Dict[str, object]
    nuc2: Dict[str, object]
    prot: Dict[str, object]

    @classmethod
    def default(cls) -> "PipelineParameters":
        """Create new default clustering parameters.

        Returns:
            `~igua.pipeline.PipelineParameters`: A new pipeline parameters
            object with all parameters initialized to defaults.

        """
        return cls(
            nuc1=dict(
                e_value=0.001,
                sequence_identity=0.85,
                coverage=1.0,
                cluster_mode=0,
                coverage_mode=1,
                spaced_kmer_mode=0,
            ),
            nuc2=dict(
                e_value=0.001,
                sequence_identity=0.6,
                coverage=0.5,
                cluster_mode=0,
                coverage_mode=0,
                spaced_kmer_mode=0,
            ),
            prot=dict(
                e_value=0.001,
                coverage=0.9,
                coverage_mode=1,
                sequence_identity=0.5,
            ),
        )


@dataclasses.dataclass
class PipelineResult:
    """The results of the IGUA clustering pipeline.

    Attributes:
        gcfs (`pandas.DataFrame`): A dataframe with columns ``cluster_id``,
            ``cluster_length``, ``gcf_id``, ``gcf_representative``,
            ``nucleotide_representative``, ``fragment_representative``
            and ``filename`` summarizing the generated gene cluster families
            at each stage.
        compositions (`anndata.AnnData` or `None`): The protein compositions
            of the stage 2 representative clusters generated by IGUA. May
            be `None` if no clustering method was requested, or if stage 2
            already only produced a singleton.

    """
    gcfs: pandas.DataFrame
    compositions: typing.Optional[anndata.AnnData]


class Pipeline:
    """The IGUA multi-stage clustering pipeline.
    """

    def __init__(
        self,
        strategy: ClusteringStrategy = default_strategy(),
        params: typing.Optional[PipelineParameters] = None,
        *,
        prefix: str = "GCF",
        jobs: int = 1,
        weight: Literal["protein", None] = "protein",
        mmseqs: typing.Optional[MMSeqs] = None,
        progress: typing.Optional[rich.progress.Progress] = None,
        workdir: typing.Optional[pathlib.Path] = None,
    ):
        """Create a new pipeline.

        Arguments:
            strategy (`ClusteringStrategy` or `None`): The clustering
                strategy to use to form the final clusters using the
                protein compositions in the stage 3 of the pipeline.
                Pass `None` to disable the stage 3 altogether.
            params (`PipelineParameters` or `None`): An object with
                parameters to pass to MMseqs2 for each stages of the
                pipeline.
            prefix (`str`): The prefix to use for the GCF identifiers.
            jobs (`int`): The number of threads to use with MMseqs2 and
                the distance computation in stage 3.
            weight (`str` or `None`): How to weight dimensions in the
                protein composition matrix. Use `protein` to weigh by
                protein length, or `None` to disable weighting.
            mmseqs (`~igua.mmseqs.MMseqs`): The MMSeqs2 driver to use
                to conduct linear clustering in stages 1 and 2 and
                protein clustering in stage 3.
            progress (`~rich.progress.Progress` or `None`): A `rich`
                progress object to use for reporting progress.
            workdir (`pathlib.Path` or `None`): The path to a folder
                to use for temporary files. If `None` given, generates
                a temporary folder with `tempfile.TemporaryDirectory`
                on each `~Pipeline.run` invokation.

        """
        self.jobs = jobs
        self.params = params or PipelineParameters.default()
        self.workdir = None if workdir is None else pathlib.Path(workdir)
        self.prefix = prefix
        self.mmseqs = mmseqs

        self.progress = progress
        if progress is None:
            self.console = rich.console.Console(quiet=True)
        else:
            self.console = progress.console

        if weight is not None and weight != "protein":
            raise ValueError(f"invalid weight: {weight!r}")
        self.weight = weight

        if strategy is None:
            self.clustering_strategy = None
        else:
            if not isinstance(strategy, ClusteringStrategy):
                raise TypeError(f"expected ClusteringStrategy, got {type(strategy).__name__}")
            self.clustering_strategy = copy.deepcopy(strategy)
            self.clustering_strategy.jobs = self.jobs

    # ---

    def _extract_clusters_to_file(
        self,
        dataset: BaseDataset,
        output: pathlib.Path,
    ) -> pandas.DataFrame:
        """Extract the dataset clusters to the given file.

        Arguments:
            dataset (`~igua.dataset.base.BaseDataset`): The dataset
                containing the gene clusters to extract.
            output (`pathlib.Path`): The path to the output file to
                write the gene clusters sequence to.

        Returns:
            `pandas.DataFrame`: A table with columns ``cluster_id`` and
            ``cluster_length`` reporting the ID and lengths of the
            extracted clusters.

        """
        # record processed sequences to flag duplicates
        done = set()
        n_duplicates = 0
        # extract raw sequences
        self.console.print(f"[bold blue]{'Loading':>12}[/] input clusters")
        with output.open("w") as dst:
            sink = _FASTASink(dst)
            for cluster in dataset.extract_clusters(self.progress):
                if cluster.id not in done:
                    done.add(cluster.id)
                    sink.add_record(cluster.id, cluster.sequence, filename=cluster.source)
                else:
                    n_duplicates += 1
            input_sequences = (
                sink.report_statistic()
                    .rename(columns={"id": "cluster_id", "length": "cluster_length"})
                    .set_index("cluster_id")
            )
        if n_duplicates > 0:
            self.console.print(
                f"[bold yellow]{'Warning':>12}[/] {n_duplicates} duplicate "
                "clusters found in inputs"
            )

        self.console.print(
            f"[bold green]{'Loaded':>12}[/] {len(input_sequences)} "
            "clusters to process"
        )
        return input_sequences

    def _extract_proteins_to_file(
        self,
        dataset: BaseDataset,
        clusters: typing.Collection[str],
        output: pathlib.Path,
    ) -> pandas.DataFrame:
        """Extract the dataset proteins to the given file.

        Arguments:
            dataset (`~igua.dataset.BaseDataset`): The dataset containing
                the gene clusters to extract and their proteins.
            clusters (`~collections.abc.Collection` of `str`): A collection
                of clusters from which to extract proteins (typically the
                clusters selected as representatives in the previous step).
            output (`pathlib.Path`): The path to the output file to write
                proteins to.

        Returns:
            `pandas.DataFrame`: A table with columns ``protein_id``,
            ``protein_length`` and ``cluster_id`` reporting the ID, length
            and source cluster of the extracted proteins.

        """
        self.console.print(
            f"[bold blue]{'Extracting':>12}[/] protein sequences from "
            "representative clusters"
        )
        with output.open("w") as dst:
            sink = _FASTASink(dst)
            for protein in dataset.extract_proteins(self.progress, clusters):
                sink.add_record(protein.id, protein.sequence, cluster_id=protein.cluster_id)
            protein_sizes = (
                sink.report_statistic()
                    .rename(columns={"id": "protein_id", "length": "protein_length"})
                    .set_index("protein_id")
                    .astype({"protein_length": "int32"})
            )
        return protein_sizes

    def _make_compositions(
        self,
        protein_clusters: pandas.DataFrame,
        representatives: typing.Dict[str, int],
        protein_representatives: typing.Dict[str, int],
    ) -> anndata.AnnData:
        """Build a compositional matrix from the given protein clusters.

        Arguments:
            protein_clusters (`pandas.DataFrame`): The clustering of
                proteins returned by MMSeqs.
            representatives (`dict` of `str` to `int`): The index mapping
                gene cluster IDs to their index in the compositional table.
            protein_representatives (`dict` of `str` to `int`): The index
                mapping protein IDs to their index in the compositional
                table.

        Returns:
            `~anndata.AnnData`: An annotated compositional matrix where
            each cluster is represented as a combination of protein
            counts.

        """
        self.console.print(
            f"[bold blue]{'Building':>12}[/] protein compositional array"
        )

        # index the proteins by protein ID
        lengths = protein_clusters.set_index("protein_id")["protein_length"]

        # create the empty compositional matrix
        compositions = scipy.sparse.dok_matrix(
            (len(representatives), len(protein_representatives)),
            dtype=numpy.int32
        )

        # Build compositional matrix using protein lengths as weights
        task = self.progress.add_task(
            description=f"[bold blue]{'Working':>9}[/]",
            total=len(protein_clusters)
        )
        for row in self.progress.track(protein_clusters.itertuples(), task_id=task):
            cluster_index = representatives[row.cluster_id]
            prot_index = protein_representatives[row.protein_representative]
            compositions[cluster_index, prot_index] += 1 #lengths.loc[row.protein_representative]
        self.progress.remove_task(task)

        # Build the annotated data matrix
        sorted_representatives = sorted(representatives, key=representatives.__getitem__)
        sorted_protein_representatives = sorted(
            protein_representatives, key=protein_representatives.__getitem__
        )
        return anndata.AnnData(
            X=compositions.tocsr(),
            obs=pandas.DataFrame(
                index=pandas.Index(sorted_representatives, name="cluster_id")
            ),
            var=pandas.DataFrame(
                index=pandas.Index(sorted_protein_representatives, name="protein_id"),
                data=dict(size=lengths.loc[sorted_protein_representatives]),
            ),
        )

    def _run_nucleotide_deduplication(
        self,
        workdir: pathlib.Path,
        clusters_fna: pathlib.Path,
        mmseqs: MMSeqs,
    ) -> Tuple[Clustering, pandas.DataFrame]:
        """Run the nucleotide deduplication stage of the pipeline.

        Arguments:
            workdir (`pathlib.Path`): The path to the temporary directory
                to use for processing data.
            clusters_fna (`pathlib.Path`): The path to a two-line FASTA
                file containing the gene clusters to process.

        Returns:
            (`~igua.mmseqs.Clustering`, `pandas.DataFrame`): The clustering
            file created by MMseqs, and a dataframe containing a summary of
            the result.

        """
        # create initial sequence database
        self.console.print(
            f"[bold blue]{'Starting':>12}[/] nucleotide deduplication step "
            "with [purple]MMSeqs2[/]"
        )
        with Stopwatch() as stopwatch:
            db = Database.create(mmseqs, clusters_fna)
            step1 = db.cluster(workdir / "step1.db", **self.params.nuc1)
            gcfs1 = step1.to_dataframe(columns=["fragment_representative", "cluster_id"]).sort_values("cluster_id")  # type: ignore
        self.console.print(
            f"[bold blue]{'Finished':>12}[/] nucleotide deduplication step "
            f"in [bold cyan]{stopwatch.total_human()}[/]"
        )
        self.console.print(
            f"[bold green]{'Reduced':>12}[/] {len(gcfs1)} clusters to "
            f"{gcfs1.fragment_representative.nunique()} complete representatives"
        )
        return step1, gcfs1

    def _run_nucleotide_clustering(
        self,
        workdir: pathlib.Path,
        step1: Clustering,
    ) -> pandas.DataFrame:
        """Run the nucleotide clustering stage of the pipeline.

        Arguments:
            workdir (`pathlib.Path`): The path to the temporary directory
                to use for processing data.
            step1 (`~igua.mmseqs.Clustering`): The MMseqs clustering file
                created at the previous stage.

        Returns:
            `pandas.DataFrame`: The result of the clustering on the
            representative sequences of the ``step1`` clustering.

        """
        # cluster sequences
        self.console.print(
            f"[bold blue]{'Starting':>12}[/] nucleotide clustering step with [purple]MMSeqs2[/]"
        )
        with Stopwatch() as stopwatch:
            repdb = step1.to_subdb(workdir / "step1.rep_seq.db")
            step2 = repdb.cluster(workdir / "step2.db", **self.params.nuc2)
            gcfs2 = step2.to_dataframe(columns=["nucleotide_representative", "fragment_representative"]).sort_values("fragment_representative")  # type: ignore
        self.console.print(
            f"[bold blue]{'Finished':>12}[/] nucleotide clustering step "
            f"in [bold cyan]{stopwatch.total_human()}[/]"
        )
        self.console.print(
            f"[bold green]{'Reduced':>12}[/] {len(gcfs2)} clusters to "
            f"{len(gcfs2.nucleotide_representative.unique())} nucleotide "
            "representatives"
        )
        return gcfs2

    def _extract_representatives(
        self,
        gcfs2: pandas.DataFrame
    ) -> Dict[str, int]:
        """Extract the representative clusters in ``gcfs2`` into a index.

        Arguments:
            gcfs2 (`pandas.DataFrame`): The clustering results of the
                second stage.

        Returns:
            `dict` of `str` to `int`: A dictionary mapping every unique
            representative cluster in ``gcfs2`` to a single index.

        """
        # load representatives
        self.console.print(
            f"[bold blue]{'Extracting':>12}[/] representative clusters"
        )
        representatives = {
            x: i
            for i, x in enumerate(sorted(gcfs2["nucleotide_representative"].unique()))
        }
        self.console.print(
            f"[bold green]{'Found':>12}[/] {len(representatives)} nucleotide "
            "representative clusters"
        )
        return representatives

    def _run_protein_clustering(
        self,
        workdir: pathlib.Path,
        dataset: BaseDataset,
        representatives: typing.Collection[str],
        mmseqs: MMSeqs,
    ) -> anndata.AnnData:
        """Run protein clustering and build a compositional matrix.

        Arguments:
            workdir (`pathlib.Path`): The path to the temporary directory
                to use for processing data.
            dataset (`~igua.dataset.BaseDataset`): The dataset containing
                the gene clusters to extract and their proteins.
            representatives (`collections.abc.Collection` of `str`): A
                collection of gene cluster IDs to extract proteins from.

        Returns:
            `anndata.AnnData`: A compositional matrix derived from the
            protein clustering. See `ClusteringPipeline._make_compositions`
            for more information.

        """
        # extract proteins and record sizes
        proteins_faa = workdir / "proteins.faa"
        protein_sizes = self._extract_proteins_to_file(dataset, representatives, output=proteins_faa)
        if not proteins_faa.exists() or proteins_faa.stat().st_size == 0:
            self.console.print(
                f"[bold yellow]{'Warning':>12}[/] No proteins extracted from input dataset"
            )

        # cluster proteins
        self.console.print(
            f"[bold blue]{'Starting':>12}[/] protein clustering step with [purple]MMSeqs2[/]"
        )
        with Stopwatch() as stopwatch:
            prot_db = Database.create(mmseqs, proteins_faa)
            prot_result = prot_db.cluster(workdir / "step3.db", **self.params.prot)
            prot_clusters = prot_result.to_dataframe(
                columns=["protein_representative", "protein_id"]
            )
            # record protein lengths and source cluster
            prot_clusters = pandas.merge(
                prot_clusters,
                protein_sizes[["cluster_id", "protein_length"]],
                on="protein_id"
            )
            # extract protein representatives
            protein_representatives = {
                x: i
                for i, x in enumerate(
                    sorted(prot_clusters["protein_representative"].unique())
                )
            }
        self.console.print(
            f"[bold blue]{'Finished':>12}[/] protein clustering step "
            f"in [bold cyan]{stopwatch.total_human()}[/]"
        )
        self.console.print(
            f"[bold green]{'Found':>12}[/] {len(protein_representatives)} "
            f"protein representatives for {len(prot_clusters)} proteins"
        )
        return self._make_compositions(
            prot_clusters,
            representatives,
            protein_representatives,
        )

    def _hierarchical_clustering(
        self,
        compositions: anndata.AnnData,
    ) -> pandas.DataFrame:
        """Perform hierarchical (or linear) clustering on the given data.

        Arguments:
            compositions (`anndata.AnnData`): A compositional matrix
                obtained from the protein clustering.

        Returns:
            `pandas.DataFrame`: A table mapping each row of ``compositions``
            to a GCF with arbitrary identifiers.

        """
        assert self.clustering_strategy is not None
        # use weights unless disabled
        if self.weight == "protein":
            weights = compositions.var["size"].values
        elif self.weight is None:
            weights = None
        else:
            raise ValueError(f"invalid clustering weight: {self.weight!r}")
        # run clustering based on protein array membership
        self.console.print(
            f"[bold blue]{'Clustering':>12}[/] gene clusters using "
            f"{self.clustering_strategy.method} linkage"
        )
        with Stopwatch() as stopwatch:
            flat = self.clustering_strategy.cluster(compositions.X, weights)
            # build GCFs based on flat clustering
            gcfs3 = pandas.DataFrame(
                {
                    "gcf_id": [f"{self.prefix}{i:07}" for i in flat],
                    "nucleotide_representative": compositions.obs_names,
                }
            )
        self.console.print(
            f"[bold blue]{'Finished':>12}[/] hierarchical clustering step "
            f"in [bold cyan]{stopwatch.total_human()}[/]"
        )
        return gcfs3

    def _hierarchical_clustering_fallback(
        self,
        representatives: Mapping[str, int],
    ) -> pandas.DataFrame:
        self.console.print(
            f"[bold yellow]{'Skipping':>12}[/] hierarchical clustering step"
        )
        compositions = None
        # proteins_faa = None
        sorted_representatives = sorted(
            representatives,
            key=representatives.__getitem__
        )
        gcfs3 = pandas.DataFrame(
            {
                "gcf_id": [
                    f"{self.prefix}{i+1:07}"
                    for i in range(len(sorted_representatives))
                ],
                "nucleotide_representative": sorted_representatives,
            }
        )
        return gcfs3

    def _join_results(
        self,
        gcfs1: pandas.DataFrame,
        gcfs2: pandas.DataFrame,
        gcfs3: pandas.DataFrame,
        input_sequences: pandas.DataFrame,
    ) -> pandas.DataFrame:
        """Join the results of the different stages to a single dataframe.
        """
        #
        self.console.print(
            f"[bold green]{'Built':>12}[/] {len(gcfs3.gcf_id.unique())} "
            f"GCFs from {len(input_sequences)} initial clusters"
        )

        # build GCFs based on flat clustering
        gcf3_representatives = (
            pandas.merge(
                gcfs3,
                input_sequences["cluster_length"],
                left_on="nucleotide_representative",
                right_index=True,
            )
            .sort_values("cluster_length")
            .drop_duplicates("gcf_id", keep="last")
            .set_index("gcf_id")
        )
        gcfs3 = pandas.merge(
            gcfs3,
            gcf3_representatives["nucleotide_representative"].rename(
                "gcf_representative"
            ),
            left_on="gcf_id",
            right_index=True,
        )

        # build final GCF table
        gcfs = pandas.merge(
            pandas.merge(
                pandas.merge(gcfs1, gcfs2, on="fragment_representative"),
                gcfs3,
                on="nucleotide_representative",
            ),
            input_sequences,
            left_on="cluster_id",
            right_index=True,
        )

        # sort results and drop unused columns
        gcfs.sort_values(["gcf_id", "cluster_length"], inplace=True)
        gcfs = gcfs[
            [
                "cluster_id",
                "cluster_length",
                "gcf_id",
                "gcf_representative",
                "nucleotide_representative",
                "fragment_representative",
                "filename",
            ]
        ]
        return gcfs

    # ---

    def run(
        self,
        dataset: BaseDataset,
    ):
        """Run the clustering pipeline on the given dataset.

        Arguments:
            dataset (`~igua.dataset.base.BaseDataset`): A dataset containing
                the gene clusters to cluster into families.

        Returns:
            `~igua.pipeline.PipelineResult`: The results of the pipeline.
            See class-level documentation for more information.

        """
        # create a context to use the user-provided temporary directory
        # or create one if none was given
        if self.workdir is None:
            workdir_context = tempfile.TemporaryDirectory(prefix="igua-")
        else:
            workdir_context = contextlib.nullcontext(self.workdir)

        with workdir_context as workdir:
            workdir = pathlib.Path(workdir)
            # use the user-provided MMseqs or create a default one
            mmseqs = self.mmseqs or MMSeqs(
                progress=self.progress,
                threads=self.jobs,
                tempdir=workdir
            )
            # extract input nucleotides
            clusters_fna = workdir / "clusters.fna"
            input_sequences = self._extract_clusters_to_file(dataset, clusters_fna)
            # run first two steps of clustering and get nucleotide representatives
            step1, gcfs1 = self._run_nucleotide_deduplication(workdir, clusters_fna, mmseqs)
            gcfs2 = self._run_nucleotide_clustering(workdir, step1)
            representatives = self._extract_representatives(gcfs2)
            # run protein clustering if enabled and if there are enough
            # gene clusters left
            if self.clustering_strategy is not None and len(representatives) > 1:
                compositions = self._run_protein_clustering(
                    workdir,
                    dataset,
                    representatives,
                    mmseqs,
                )
                gcfs3 = self._hierarchical_clustering(
                    compositions,
                )
            else:
                compositions = None
                gcfs3 = self._hierarchical_clustering_fallback(
                    representatives
                )

        # generate result table
        gcfs = self._join_results(gcfs1, gcfs2, gcfs3, input_sequences)
        return PipelineResult(gcfs=gcfs, compositions=compositions)



import pathlib
import typing
from typing import Sequence, Iterable

import pandas
import rich.progress

from .base import BaseDataset, Cluster, Protein


class DatasetList(BaseDataset, Sequence[BaseDataset]):
    """A dataset consisting in a list of other datasets.
    """

    def __init__(self, datasets: Iterable[BaseDataset] = ()):
        super().__init__()
        self.datasets = list(datasets)

    def __len__(self):
        return len(self.datasets)

    def __getitem__(self, index):
        return self.datasets[index]

    def __iter__(self):
        return iter(self.datasets)

    def extract_clusters(
        self,
        progress: rich.progress.Progress,
    ) -> typing.Iterator[Cluster]:
        task = progress.add_task(f"[bold blue]{'Working':>9}[/]")
        for dataset in progress.track(self.datasets, task_id=task):
            yield from dataset.extract_clusters(progress)
        progress.remove_task(task)

    def extract_proteins(
        self,
        progress: rich.progress.Progress,
        representatives: typing.Collection[str],
    ) -> typing.Iterator[Protein]:
        task = progress.add_task(f"[bold blue]{'Working':>9}[/]")
        for dataset in progress.track(self.datasets, task_id=task):
            yield from dataset.extract_proteins(progress, representatives)
        progress.remove_task(task)
                    


import gzip
import io
import re
import pathlib
import typing
import zipfile

import gb_io
import pandas as pd
import rich.progress

from .base import BaseDataset, Cluster, Protein

_GZIP_MAGIC = b"\x1f\x8b"


def _extract_clusters_from_record(
    record: gb_io.Record,
    source: str,
) -> typing.Iterable[Cluster]:
    for region in filter(lambda feat: feat.kind == "region", record.features):
        start = region.location.start
        end = region.location.end
        number = next(q.value for q in region.qualifiers if q.key == "region_number")
        region_id = f"{record.name}.region{number:>03}"
        yield Cluster(
            id=region_id,
            sequence=record.sequence[start-1:end].decode('ascii'),
            source=source,
        )


def _extract_proteins_from_record(
    record: gb_io.Record,
    cluster_ids: typing.Container[str],
) -> typing.Iterable[Protein]:
    for region in filter(lambda feat: feat.kind == "region", record.features):
        number = next(q.value for q in region.qualifiers if q.key == "region_number")
        region_id = f"{record.name}.region{number:>03}"
        if region_id not in cluster_ids:
            continue
        for cds in filter(
            lambda feat: feat.kind == "CDS" and feat.location.start >= region.location.start and feat.location.end <= region.location.end,
            record.features,
        ):
            locus_tag = next(q.value for q in cds.qualifiers if q.key == "locus_tag")
            translation = next(q.value for q in cds.qualifiers if q.key == "translation")
            yield Protein(
                f"{region_id}_{locus_tag}",
                translation,
                cluster_id=region_id
            )


class AntiSMASHGenBankDataset(BaseDataset):
    """A dataset composed of antiSMASH regions in a GenBank file.

    antiSMASH reports regions in a GenBank file but not necessarily
    with one region per record. This class supports extracting regions
    independently and recovering the right cluster ID per region.

    """

    def __init__(
        self,
        path: pathlib.Path,
        *,
        mode: str = "region",
    ):
        super().__init__()

        if mode not in {"region", "protocluster", "core"}:
            raise ValueError(f"invalid mode: {mode!r}")
        if mode in {"protocluster", "core"}:
            raise NotImplementedError(f"not implemented mode: {mode!r}")

        self.mode = mode
        self.path = path

    def extract_clusters(
        self,
        progress: rich.progress.Progress,
    ) -> typing.Iterable[Cluster]:
        task = progress.add_task(f"[bold blue]{'Reading':>9}[/]")
        with io.BufferedReader(progress.open(self.path, "rb", task_id=task)) as reader:  # type: ignore
            if reader.peek().startswith(_GZIP_MAGIC):
                reader = gzip.GzipFile(mode="rb", fileobj=reader)  # type: ignore
            for record in gb_io.iter(reader):
                yield from _extract_clusters_from_record(record, source=str(self.path))
        progress.remove_task(task)

    def extract_proteins(
        self,
        progress: rich.progress.Progress,
        cluster_ids: typing.Container[str],
    ) -> typing.Iterable[Protein]:
        task = progress.add_task(f"[bold blue]{'Reading':>9}[/]")
        with io.BufferedReader(progress.open(self.path, "rb", task_id=task)) as reader:  # type: ignore
            if reader.peek()[:2] == b"\x1f\x8b":
                reader = gzip.GzipFile(mode="rb", fileobj=reader)  # type: ignore
            for record in gb_io.iter(reader):
                yield from _extract_proteins_from_record(record, cluster_ids)
        progress.remove_task(task)


class AntiSMASHZipDataset(BaseDataset):
    """A dataset composed of antiSMASH results in a Zip file.

    antiSMASH can be configured to report all results inside a Zip file.
    This class supports reading the antiSMASH-predicted regions from a
    Zip archive, including handling of region IDs, without requiring the
    archive to be decompressed.

    """

    _REGION_RX = re.compile(r"region(\d){3,}.gbk$")

    def __init__(
        self,
        path: pathlib.Path,
        *,
        mode: str = "region",
    ):
        super().__init__()

        if mode not in {"region", "protocluster", "core"}:
            raise ValueError(f"invalid mode: {mode!r}")
        if mode in {"protocluster", "core"}:
            raise NotImplementedError(f"not implemented mode: {mode!r}")

        self.mode = mode
        self.path = path

    def extract_clusters(
        self,
        progress: rich.progress.Progress,
    ) -> typing.Iterable[Cluster]:
        with zipfile.ZipFile(self.path, mode="r") as archive:
            for file in archive.filelist:
                if self._REGION_RX.search(file.filename) is None:
                    continue
                with archive.open(file, mode="r") as src:
                    records = gb_io.iter(src)
                    record = next(records)
                    if next(records, None) is not None:
                        raise ValueError(f"more than one record found in {file.filename}")
                    yield from _extract_clusters_from_record(record, source=str(self.path))

    def extract_proteins(
        self,
        progress: rich.progress.Progress,
        cluster_ids: typing.Container[str],
    ) -> typing.Iterable[Protein]:
        with zipfile.ZipFile(self.path, mode="r") as archive:
            files = { file.filename for file in archive.filelist }
            for file in archive.filelist:
                if self._REGION_RX.search(file.filename) is None:
                    continue
                with archive.open(file, mode="r") as src:
                    records = gb_io.iter(src)
                    record = next(records)
                    if next(records, None) is not None:
                        raise ValueError(f"more than one record found in {file.filename}")
                    yield from _extract_proteins_from_record(record, cluster_ids)




import argparse
import csv
import os
import io
import itertools
import tarfile
import glob
import re
import multiprocessing.pool
import sys

import Bio.SeqIO
import gecco.model
import gecco.hmmer
import gecco.crf
import pandas
import numpy
import scipy.sparse
import rich
import rich.progress
import tqdm


class TrackedFile(io.RawIOBase):

    def __init__(self, handle, progress, task, scale=0):
        self.handle = handle
        self.progress = progress
        self.task = task
        self.scale = scale

    def __enter__(self):
        self.handle.__enter__()
        return self

    def __exit__(self, exc_val, exc_ty, tb):
        self.handle.__exit__(exc_val, exc_ty, tb)
        return False

    def readable(self):
        return True

    def read(self, size=-1):
        block = self.handle.read(size)
        self.progress.update(self.task, advance=len(block) / (1024 ** self.scale))
        return block

    def close(self):
        self.handle.close()



GBK_RX = re.compile(r".*region(\d{3})\.gbk$")

RECORDS = [
    "proG3_total_final.gbk",
]


parser = argparse.ArgumentParser()
parser.add_argument("--input", required=True)
parser.add_argument("--output", required=True)
parser.add_argument("--pfam", required=True)
args = parser.parse_args()

pfam = gecco.hmmer.HMM("Pfam", version="35.0", url="", path=args.pfam, size=19632, relabel_with=r"s/(PF\d+).\d+/\1/")

HMMS = [ pfam ]


with rich.progress.Progress(
        rich.progress.SpinnerColumn(finished_text="[green]:heavy_check_mark:[/]"),
        "[progress.description]{task.description}",
        rich.progress.BarColumn(bar_width=60),
        rich.progress.DownloadColumn(),
        "[progress.percentage]{task.percentage:>3.0f}%",
        rich.progress.TimeElapsedColumn(),
        rich.progress.TimeRemainingColumn(),
) as progress:

    progress.console.print("[bold green]{:>12}[/] GECCO internal CRF".format("Loading"))
    crf = gecco.crf.ClusterCRF.trained()

    progress.console.print("[bold green]{:>12}[/] GCF representatives from {!r}".format("Loading", args.input))
    gcfs = pandas.read_table(args.input, comment="#")
    representative_ids = set(gcfs.gcf_representative)
    bgc_to_gcf_index = {row.gcf_representative:row.gcf_id for row in gcfs.itertuples()}

    # Extract clusters from antiSMASH results

    def region_to_cluster(record, bgc_id):
        genes = []
        for cds in filter(lambda f: f.type == "CDS", record.features):
            qualifiers = cds.qualifiers
            if "locus_tag" in qualifiers:
                id_ = qualifiers["locus_tag"][0]
            elif "protein_id" in qualifiers:
                id_ = qualifiers["protein_id"][0]
            elif "gene" in qualifiers:
                id_ = qualifiers["gene"][0]
            else:
                progress.print(cds); exit(1)
            seq = Bio.Seq.Seq(qualifiers['translation'][0])
            protein = gecco.model.Protein(id_, seq)
            gene = gecco.model.Gene(source=record, start=cds.location.start, end=cds.location.end, protein=protein, strand=gecco.model.Strand.Coding)
            genes.append(gene)
        return gecco.model.Cluster(bgc_id, genes)

    def record_to_cluster(record):
        genes = []
        for cds in filter(lambda f: f.type == "CDS", record.features):
            qualifiers = cds.qualifiers
            protein = gecco.model.Protein(record.name, Bio.Seq.Seq(qualifiers['translation'][0]))
            gene = gecco.model.Gene(record, start=cds.location.start, end=cds.location.end, protein=protein, strand=gecco.model.Strand.Coding)
            genes.append(gene)
        return gecco.model.Cluster(record.name, genes)

    clusters = []
    for filename in RECORDS:
        progress.console.print("[bold green]{:>12}[/] representatives from {!r}".format("Extracting", filename))
        task = progress.add_task("Reading")
        with progress.open(filename, "rb", task_id=task) as f:
            for record in Bio.SeqIO.parse(filename, "genbank"):
                if record.name.startswith("BGC"):
                    id_ = record.name.rsplit(".", 1)[0]
                    if id_ in representative_ids:
                        clusters.append(region_to_cluster(record, id_))
                elif record.name in representative_ids:
                    clusters.append(record_to_cluster(record))

with rich.progress.Progress(
        rich.progress.SpinnerColumn(finished_text="[green]:heavy_check_mark:[/]"),
        "[progress.description]{task.description}",
        rich.progress.BarColumn(bar_width=60),
        "[progress.completed]{task.completed:.1f}/{task.total:.1f}",
        "[progress.completed]{task.fields[unit]}",
        "[progress.percentage]{task.percentage:>3.0f}%",
        rich.progress.TimeElapsedColumn(),
        rich.progress.TimeRemainingColumn(),
) as progress:


    # Annotate with Pfam

    progress.console.print("[bold green]{:>12}[/] genes of {} BGCs".format("Annotating", len(clusters)))
    genes = [gene for cluster in clusters for gene in cluster.genes]
    for hmm in HMMS:
        hmmer = gecco.hmmer.PyHMMER(hmm, cpus=28)
        task = progress.add_task(hmm.id, total=hmm.size, unit="HMM")
        hmmer.run(genes, progress=lambda h,i: progress.update(task, advance=1))

    progress.console.print("[bold green]{:>12}[/] domains by p-value under 1e-5".format("Filtering"))
    genes = [
        gene.with_protein(gene.protein.with_domains([d for d in gene.protein.domains if d.pvalue < 1e-5]))
        for gene in genes
    ]

    progress.console.print("[bold green]{:>12}[/] genes by source BGC".format("Sorting"))
    genes.sort(key=lambda g: g.source.name)

    progress.console.print("[bold green]{:>12}[/] genes into source BGCs".format("Grouping"))
    all_clusters = []
    for bgc_id, cluster_genes in itertools.groupby(genes, lambda g: g.source.name):
        id_ = bgc_id.rsplit(".", 1)[0] if bgc_id.startswith("BGC") else bgc_id
        cluster = gecco.model.Cluster(id=id_, genes=list(cluster_genes))
        all_clusters.append(cluster)


    # --- Finalize

    progress.console.print("[bold green]{:>12}[/] {} clusters by GCF ID".format("Sorting", len(all_clusters)))
    all_clusters.sort(key=lambda cluster: bgc_to_gcf_index[cluster.id])

    progress.console.print("[bold green]{:>12}[/] all possible domains from clusters".format("Extracting"))
    all_possible = sorted({domain.name for cluster in all_clusters for gene in cluster.genes for domain in gene.protein.domains})

    os.makedirs(args.output, exist_ok=True)
    progress.console.print("[bold green]{:>12}[/] labels and domain compositions".format("Writing"))
    with open(os.path.join(args.output, "labels.tsv"), "w") as f:
        for cluster in all_clusters:
            f.write(cluster.id)
            f.write("\t")
            f.write(bgc_to_gcf_index[cluster.id])
            f.write("\n")

    with open(os.path.join(args.output, "domains.tsv"), "w") as f:
        for domain in all_possible:
            f.write(domain)
            f.write("\n")

    comp = numpy.array([c.domain_composition(all_possible) for c in all_clusters])
    scipy.sparse.save_npz(os.path.join(args.output, "compositions.npz"), scipy.sparse.coo_matrix(comp))

    # --- Check missing
    all_clusters_set = { cluster.id for cluster in all_clusters }
    progress.console.print("[bold red]{:>12}[/] compositions for {} BGCs".format("Missing", len(representative_ids.difference(all_clusters_set))))
    #for bgc_id in representative_ids.difference(all_clusters_set):
    #    progress.console.print("[bold red]{:>12}[/] compositions for {}".format("Missing", bgc_id))



import enum
import os
import pathlib
import subprocess
import typing

import pandas
import rich.progress
import tempfile


COMMANDS = {
  "easy-search",
  "easy-linsearch",
  "easy-cluster",
  "easy-linclust",
  "easy-taxonomy",
  "easy-rbh",
  "search",
  "linsearch",
  "map",
  "rbh",
  "linclust",
  "cluster",
  "clusterupdate",
  "taxonomy",
  "databases",
  "createdb",
  "createindex",
  "createlinindex",
  "convertmsa",
  "tsv2db",
  "tar2db",
  "msa2profile",
  "compress",
  "decompress",
  "rmdb",
  "mvdb",
  "cpdb",
  "lndb",
  "unpackdb",
  "touchdb",
  "createsubdb",
  "concatdbs",
  "mergedbs",
  "subtractdbs",
  "convertalis",
  "createtsv",
  "convert2fasta",
  "result2flat",
  "createseqfiledb",
  "taxonomyreport",
  "extractorfs",
  "extractframes",
  "orftocontig",
  "reverseseq",
  "translatenucs",
  "translateaa",
  "splitsequence",
  "masksequence",
  "extractalignedregion",
  "swapresults",
  "result2rbh",
  "result2msa",
  "result2dnamsa",
  "result2stats",
  "filterresult",
  "offsetalignment",
  "proteinaln2nucl",
  "result2repseq",
  "sortresult",
  "summarizealis",
  "summarizeresult",
  "createtaxdb",
  "createbintaxonomy",
  "addtaxonomy",
  "taxonomyreport",
  "filtertaxdb",
  "filtertaxseqdb",
  "aggregatetax",
  "aggregatetaxweights",
  "lcaalign",
  "lca",
  "majoritylca",
  "multihitdb",
  "multihitsearch",
  "besthitperset",
  "combinepvalperset",
  "mergeresultsbyset",
  "prefilter",
  "ungappedprefilter",
  "kmermatcher",
  "kmersearch",
  "align",
  "alignall",
  "transitivealign",
  "rescorediagonal",
  "alignbykmer",
  "clust",
  "clusthash",
  "mergeclusters",
  "result2profile",
  "msa2result",
  "msa2profile",
  "profile2pssm",
  "profile2consensus",
  "profile2repseq",
  "convertprofiledb",
  "enrich",
  "result2pp",
  "profile2cs",
  "convertca3m",
  "expandaln",
  "expand2profile",
  "view",
  "apply",
  "filterdb",
  "swapdb",
  "prefixid",
  "suffixid",
  "renamedbkeys",
  "diffseqdbs",
  "summarizetabs",
  "gff2db",
  "maskbygff",
  "convertkb",
  "summarizeheaders",
  "nrtotaxmapping",
  "extractdomains",
  "countkmer",
}


class MMSeqs(object):
    """A wrapper around an ``mmseqs`` binary and common parameters.
    """

    def __init__(
        self,
        binary: str = "mmseqs",
        progress: typing.Optional[rich.progress.Progress] = None,
        threads: typing.Optional[int] = None,
        tempdir: typing.Optional[str] = None,
    ) -> None:
        self.binary = binary
        self.progress = progress
        self.threads = threads
        self.tempdir = tempdir or tempfile.TemporaryDirectory().name

    def _wrap_progress(
        self,
        process: subprocess.Popen,
    ) -> int:
        """Wrap the progress output from ``mmseqs`` into a `rich` progress bar.
        """
        assert self.progress is not None
        assert process.stdout is not None

        buffer = bytearray()
        command = ""
        task = self.progress.add_task(f"[bold blue]{'Running':>9}[/] [purple]{command}[/]", total=65)
        bar_column = next(c for c in self.progress.columns if isinstance(c, rich.progress.BarColumn))

        for x in iter(lambda: process.stdout.read(1), b""):
            buffer.append(ord(x))
            if buffer.startswith(b"["):
                # update progress
                self.progress.update(task_id=task, completed=buffer.count(b'='))
            if buffer.endswith(b"\n"):
                # extract current command being run
                _command = next(iter(buffer.split()), b"").decode()
                if _command in COMMANDS:
                    command = _command
                    bar_column.bar_width = 40 - len(command)
                    self.progress.reset(task_id=task, description=f"[bold blue]{'Running':>9}[/] [purple]{command}[/]")
                # clear current buffer
                buffer.clear()

        self.progress.update(task_id=task, visible=False)
        self.progress.remove_task(task)
        return process.wait()

    def run(
        self,
        command: str,
        *args: typing.Union[str, bytes, os.PathLike],
        **kwargs: object
    ) -> subprocess.CompletedProcess:
        """Run an arbitrary ``mmseqs`` command.
        """
        # build actual command line
        params = [self.binary, command, *args]
        for k, v in kwargs.items():
            dash = "-" if len(k) == 1 else "--"
            flag = "".join([dash, k.replace("_", "-")])
            params.append(flag)
            params.append(repr(v))

        # use fixed number of threads
        if self.threads is not None and command in {"linclust", "search"}:
            params.extend(["--threads", str(self.threads)])

        # start mmseqs subprocess
        process = subprocess.Popen(params, stdout=subprocess.PIPE, bufsize=0)

        # wrap progress if a rich progress bar is available
        if self.progress:
            self._wrap_progress(process)
            stdout = stderr = None
        else:
            stdout, stderr = process.communicate()

        # return a completed process instance for compatibility
        return subprocess.CompletedProcess(
            params,
            process.returncode,
        )

    def version(self) -> str:
        try:
            params = [self.binary, "version"]
            process = subprocess.Popen(params, stdout=subprocess.PIPE)
            stdout, stderr = process.communicate()
            return stdout.decode().strip()
        except OSError as err:
            raise RuntimeError(f"Failed to find MMSeqs2 binary {self.binary!r}") from err


class DatabaseType(enum.IntEnum):
    Protein = 1
    Nucleotide = 2


class _MMSeqsFile(object):

    def __init__(self, mmseqs: MMSeqs, path: pathlib.Path):
        self.mmseqs = mmseqs
        self.path = path

    def remove(self):
        """Remove the MMSeqs2 file.
        """
        self.mmseqs.run("rmdb", self.path)


class Database(_MMSeqsFile):

    @classmethod
    def create(
        cls,
        mmseqs: MMSeqs,
        sequences: pathlib.Path,
        path: typing.Optional[pathlib.Path] = None,
    ) -> "Database":
        """Create a new database using a FASTA-formatted sequence file.
        """
        if path is None:
            path = sequences.with_suffix(".db")
        mmseqs.run(
            "createdb",
            sequences,
            path,
            shuffle=0,
            createdb_mode=1,
            write_lookup=0,
            id_offset=0,
            compressed=0,
        ).check_returncode()
        return cls(mmseqs, path)

    def __init__(self, mmseqs: MMSeqs, path: pathlib.Path):
        """Open a database at the given location.
        """
        super().__init__(mmseqs, path)

    def to_fasta(self, path: pathlib.Path) -> pathlib.Path:
        """Convert the sequence database to a two-line FASTA file.
        """
        self.mmseqs.run("convert2fasta", self.path, path).check_returncode()
        return path

    def cluster(
        self,
        output: pathlib.Path,
        *,
        e_value=0.001,
        sequence_identity=0.85,
        coverage=1,
        cluster_mode=0,
        coverage_mode=0,
        spaced_kmer_mode=0,
    ) -> "Clustering":
        """Run linear clustering on the database.
        """
        # run clustering
        self.mmseqs.run(
            "linclust",
            self.path,
            output,
            self.mmseqs.tempdir,
            e=e_value,
            min_seq_id=sequence_identity,
            c=coverage,
            cluster_mode=cluster_mode,
            cov_mode=coverage_mode,
            spaced_kmer_mode=spaced_kmer_mode,
            remove_tmp_files=1,
        ).check_returncode()
        return Clustering(self.mmseqs, output, self)


class Clustering(_MMSeqsFile):

    def __init__(
        self,
        mmseqs: MMSeqs,
        path: pathlib.Path,
        database: Database
    ):
        super().__init__(mmseqs, path)
        self.database = database

    def to_dataframe(self, columns: typing.Tuple[str, str]) -> pandas.DataFrame:
        """Obtain the clustering results as a table.
        """
        with tempfile.NamedTemporaryFile(
            "w",
            suffix=".tsv",
            dir=self.mmseqs.tempdir
        ) as tmp:
            self.mmseqs.run(
                "createtsv",
                self.database.path,
                self.database.path,
                self.path,
                tmp.name,
            ).check_returncode()
            return pandas.read_csv(
                tmp.name,
                sep="\t",
                header=None,
                names=list(columns),
            )

    def to_subdb(self, path: pathlib.Path) -> Database:
        """Generate a sub-database with the representatives of this clustering.
        """
        self.mmseqs.run(
            "createsubdb",
            self.path,
            self.database.path,
            path,
            subdb_mode=1,
        ).check_returncode()
        return Database(self.mmseqs, path)


import unittest

import numpy
import scipy.sparse
import scipy.spatial.distance

from igua.hca import manhattan, manhattan_pair


class TestManhattan(unittest.TestCase):

    def test_unweighted(self):
        rng = numpy.random.Generator(numpy.random.MT19937(42))
        for i in range(20):
            # create a somewhat sparse matrix 
            X = rng.integers(1, 100, size=(100, 30), dtype="int32")
            X *= (rng.random(size=X.shape) < 0.2)
            S = scipy.sparse.csr_matrix(X)
            # compute expected pdist with scipy
            d_exp = scipy.spatial.distance.pdist(X, metric="cityblock")
            # compute result with custom sparse implementation
            d_out = numpy.zeros(d_exp.shape, dtype=numpy.float64)
            manhattan(S.data, S.indices, S.indptr, None, d_out)
            # ensure equality
            self.assertEqual(d_exp.tolist(), d_out.tolist())

    def test_weighted(self):
        rng = numpy.random.Generator(numpy.random.MT19937(42))
        for i in range(20):
            # create a somewhat sparse matrix 
            X = rng.integers(1, 100, size=(100, 30), dtype="int32")
            X *= (rng.random(size=X.shape) < 0.2)
            S = scipy.sparse.csr_matrix(X)
            # generate random weights
            weights = rng.integers(1, 10, X.shape[1], dtype="int32")
            # compute expected pdist with scipy
            d_exp = scipy.spatial.distance.pdist(X, metric="cityblock", w=weights)
            # compute result with custom sparse implementation
            d_out = numpy.zeros(d_exp.shape, dtype=numpy.float64)
            manhattan(S.data, S.indices, S.indptr, weights, d_out)
            # ensure equality
            self.assertEqual(d_exp.tolist(), d_out.tolist())
            # compute result as if unweighted w/ manually applied weights
            X2 = X * weights
            S2 = scipy.sparse.csr_matrix(X2, dtype="int32")
            # compute using non-weighted implementation and weighted matrix
            d_out2 = numpy.zeros(d_exp.shape, dtype=numpy.float64)
            manhattan(S2.data, S2.indices, S2.indptr, None, d_out2)
            # ensure equality
            self.assertEqual(d_out.tolist(), d_out2.tolist())



import gzip
import io
import pathlib
import typing

import gb_io
import Bio.Seq
import pandas as pd
import rich.progress

from .base import BaseDataset, Cluster, Protein
from .._utils import zopen


class GenBankDataset(BaseDataset):
    """A dataset composed of gene clusters in a GenBank file.

    GenBank files are commonly used to distribute a genomic sequence
    with associated metadata inside a single file.

    Note:
        This method treats each GenBank record as an independent gene
        cluster and extracts the full record sequence and all the
        annotated genes. For GenBank files obtained with antiSMASH,
        please use the `~igua.dataset.antismash.AntiSMASHGenBankDataset` 
        class to enable correct processing of the antiSMASH regions.

    """

    def __init__(
        self,
        path: pathlib.Path,
        *,
        gene_feature: str = "CDS",
    ):
        """Create a new GenBank dataset.

        Args:
            input (`pathlib.Path`): The path to a GenBank file.
            gene_feature (`str`): The GenBank feature to extract gene
                sequences from. Defaults to *CDS* which is used by
                most annotation tools.

        """
        super().__init__()
        self.path = path
        self.gene_feature = gene_feature

    def extract_clusters(
        self,
        progress: rich.progress.Progress,
    ) -> typing.Iterable[Cluster]:
        task = progress.add_task(f"[bold blue]{'Reading':>9}[/]")
        with progress.open(self.path, "rb", task_id=task) as reader:  # type: ignore
            with zopen(reader) as src:
                for record in gb_io.iter(src):
                    yield Cluster(
                        id=record.name,
                        sequence=record.sequence.decode('ascii'),
                        source=str(self.path),
                    )
        progress.remove_task(task)

    def extract_proteins(
        self,
        progress: rich.progress.Progress,
        clusters: typing.Container[str],
    ) -> typing.Iterable[Protein]:
        task = progress.add_task(f"[bold blue]{'Reading':>9}[/]")
        with progress.open(self.path, "rb", task_id=task) as reader:  # type: ignore
            with zopen(reader) as src:
                for record in gb_io.iter(src):
                    if record.name in clusters:
                        for i, feat in enumerate(
                            filter(lambda f: f.kind == self.gene_feature, record.features)
                        ):
                            qualifier = next(
                                (
                                    qualifier
                                    for qualifier in feat.qualifiers
                                    if qualifier.key == "translation"
                                ),
                                None,
                            )
                            if qualifier is None:
                                rich.print(  # FIXME: use warning
                                    f"[bold yellow]{'Warning':>12}[/] no "
                                    f"'translation' qualifier found in "
                                    f"{self.gene_feature!r} feature of "
                                    f"{record.name!r}"
                                )
                                loc = feat.location
                                gene = record.sequence[loc.start : loc.end]
                                translation = self._translate_orf(gene)
                            else:
                                translation = qualifier.value.rstrip("*")
                            yield Protein(
                                id="{}_{}".format(record.name, i),
                                sequence=translation,
                                cluster_id=record.name,
                            )
        progress.remove_task(task)

    def _translate_orf(
        self, sequence: typing.Union[str, bytes], translation_table: int = 11
    ) -> str:
        seq = Bio.Seq.MutableSeq(sequence)
        rest = len(seq) % 3
        if rest != 0:
            rich.print(  # FIXME: use warning
                f"[bold yellow]{'Warning':>12}[/] CDS feature location "
                "contains an incomplete codon"
            )
            seq.extend('N' * (3 - rest))
        return str(seq.translate(translation_table))


import argparse
import pathlib
import typing

from ..dataset.base import BaseDataset
from ..dataset.genbank import GenBankDataset
from ..dataset.antismash import AntiSMASHGenBankDataset, AntiSMASHZipDataset
from ..dataset.defensefinder import DefenseFinderDataset


class BaseInput:
    
    def __init__(self, filename: str):
        self.filename = filename


class GenBankInput(BaseInput):
    
    def __init__(self, filename: str):
        self.filename = pathlib.Path(filename)

    def to_dataset(self, args: argparse.Namespace) -> BaseDataset:
        return GenBankDataset(self.filename)


class GenBankListInput(BaseInput):

    def to_dataset(self, args: argparse.Namespace) -> BaseDataset:
        datasets = []
        with open(self.filename) as files:
            for file in files:
                datasets.append(GenBankDataset(pathlib.Path(file.strip())))
        return DatasetList(datasets)


class AntiSMASHGenBankInput(BaseInput):

    def to_dataset(self, args: argparse.Namespace) -> BaseDataset:
        return AntiSMASHGenBankDataset(self.filename)


class AntiSMASHZipInput(BaseInput):

    def to_dataset(self, args: argparse.Namespace) -> BaseDataset:
        return AntiSMASHZipDataset(self.filename)


class DefenseFinderTSV(BaseInput):

    def to_dataset(self, args: argparse.Namespace) -> BaseDataset:
        return DefenseFinderDataset([self.filename])


import pathlib
import typing

import pandas
from rich.console import Console

from .fasta_gff import FastaGFFDataset


class DefenseFinderDataset(FastaGFFDataset):
    """DefenseFinder-specific dataset class."""

    def __init__(
        self,
        inputs: typing.List[pathlib.Path],
        activity_filter: str = "all",
    ) -> None:
        """Initialize DefenseFinder dataset.

        Args:
            inputs: List of input file paths (typically metadata TSV).
            activity_filter: Filter by activity type. Options: 'all',
                'defense', 'antidefense'.
        """
        defensefinder_mapping = {
            "cluster_id": "sys_id",
            "start_gene": "sys_beg",
            "end_gene": "sys_end",
            "genes_in_cluster": "protein_in_syst",
        }

        super().__init__(inputs, column_mapping=defensefinder_mapping)
        self.activity_filter = activity_filter

    def _load_and_filter_systems(
        self, tsv_path: pathlib.Path, console: Console
    ) -> pandas.DataFrame:
        """Load DefenseFinder systems with activity filtering.
        
        Args:
            tsv_path: Path to the DefenseFinder systems TSV file.
            console: Rich console for logging output.
            
        Returns:
            Filtered Polars DataFrame containing system data.
        """
        df = pandas.read_csv(tsv_path, sep="\t")
        original_count = len(df)

        if self.activity_filter.lower() != "all":
            if "activity" in df.columns:
                df = df[
                    df["activity"].str.lower() == self.activity_filter.lower()
                ]
                console.print(
                    f"[bold green]{'Filtered':>12}[/] {original_count} → {len(df)} "
                    f"([bold cyan]{self.activity_filter}[/] only)"
                )
            else:
                console.print(
                    f"[bold yellow]{'Warning':>12}[/] No 'activity' column found"
                )

        return df


import collections
import gc
import gzip
import io
import pathlib
import pickle
import re
import tarfile
import typing
import uuid
from dataclasses import asdict, dataclass

import Bio.SeqIO
import pandas
import rich.progress
from rich.console import Console

from .base import BaseDataset, Cluster, Protein
from .._utils import zopen


class TarCache:
    """Cache for extracted tar members to avoid repeated extraction."""

    def __init__(self):
        """Initialize empty cache."""
        self._cache: typing.Dict[str, bytes] = {}
        self._tar_handles: typing.Dict[pathlib.Path, tarfile.TarFile] = {}

    def get_member(self, tar_path: pathlib.Path, member_path: str) -> io.BytesIO:
        """Get tar member, using cache if available."""
        cache_key = f"{tar_path}::{member_path}"

        if cache_key not in self._cache:
            if tar_path not in self._tar_handles:
                self._tar_handles[tar_path] = tarfile.open(tar_path, "r:*")

            tf = self._tar_handles[tar_path]
            member = tf.extractfile(member_path)
            if member is None:
                raise FileNotFoundError(f"Member {member_path} not found in {tar_path}")

            self._cache[cache_key] = member.read()
            member.close()

        return io.BytesIO(self._cache[cache_key])

    def cleanup(self):
        """Close all tar handles and clear cache."""
        for tf in self._tar_handles.values():
            tf.close()
        self._tar_handles.clear()
        self._cache.clear()


def _parse_tar_path(
    path: pathlib.Path,
) -> typing.Union[tuple[pathlib.Path, str], tuple[None, None]]:
    """Parse a path that might point inside a tar archive.

    Args:
        path: Path that might contain a tar archive reference.

    Returns:
        Tuple of (tar_path, member_path) if tar archive found, else (None, None).
    """
    path_str = str(path)

    for ext in [".tar.gz", ".tar.bz2", ".tgz", ".tar"]:
        if ext in path_str:
            parts = path_str.split(ext, 1)
            tar_path = pathlib.Path(parts[0] + ext)
            if tar_path.exists():
                member_path = parts[1].lstrip("/")
                return tar_path, member_path

    return None, None


def smart_open(
    path: pathlib.Path,
    mode: str = "rb",
    tar_cache: typing.Optional[TarCache] = None,
) -> typing.BinaryIO:
    """Open file, handling regular files and tar archives.

    Uses caching for tar members to avoid repeated extraction.
    Supports gzip compression for both file types.

    Args:
        path: Path to file (may be inside tar archive).
        mode: File mode ('rb' or 'rt').
        tar_cache: Optional TarCache instance for caching tar members.

    Returns:
        File-like object (binary mode).
    """
    tar_path, member_path = _parse_tar_path(path)

    if tar_path and member_path:
        if tar_cache is None:
            tar_cache = TarCache()
        member_data = tar_cache.get_member(tar_path, member_path)
        return zopen(member_data)
    else:
        return zopen(path)


@dataclass
class SystemCoordinates:
    """Genome coordinates for a gene cluster.

    Attributes:
        cluster_id: Cluster identifier.
        seq_id: Sequence/contig identifier.
        start_coord: Start coordinate on the sequence (1-based).
        end_coord: End coordinate on the sequence (1-based, inclusive).
        strand: Strand orientation ('+', '-', or '.').
        genes: List of gene identifiers in the system.
        fasta_file: Path to the genome FASTA file.
        valid: Whether the coordinates are valid.
        error_msg: Error message if coordinates are invalid.
    """

    cluster_id: str
    seq_id: str
    start_coord: int
    end_coord: int
    strand: str
    genes: typing.List[str]
    fasta_file: str
    valid: bool = True
    error_msg: typing.Optional[str] = None

    def to_dict(self) -> typing.Dict:
        """Convert to dictionary for serialization.

        Returns:
            Dictionary representation of the coordinates.
        """
        return asdict(self)

    @classmethod
    def from_dict(cls, data: typing.Dict) -> "SystemCoordinates":
        """Create from dictionary.

        Args:
            data: Dictionary containing coordinate data.

        Returns:
            SystemCoordinates instance.
        """
        return cls(**data)


class GFFIndex:
    """Fast, in-memory GFF index for gene feature lookup.

    Attributes:
        path: Path to the GFF file.
    """

    def __init__(
        self, gff_path: pathlib.Path, tar_cache: typing.Optional[TarCache] = None
    ):
        """Initialize GFF index.

        Args:
            gff_path: Path to the GFF file.
            tar_cache: Optional TarCache instance for caching tar members.
        """
        self.path = gff_path
        self._tar_cache = tar_cache
        self._index: typing.Dict[str, typing.Dict] = {}
        self._build_index()

    def _build_index(self):
        """Build comprehensive ID index from GFF file.

        Creates multiple lookup variants for each gene/CDS feature
        including ID, locus_tag, Name, gene, old_locus_tag, and
        protein_id attributes. Also creates prefixed variants
        (gene-, cds-) and underscore/tilde variants.
        """
        with smart_open(self.path, tar_cache=self._tar_cache) as reader:
            for line in io.TextIOWrapper(reader, encoding="utf-8"):
                if line.startswith("#") or not line.strip():
                    continue

                parts = line.strip().split("\t")
                if len(parts) < 9 or parts[2] not in ["gene", "CDS"]:
                    continue

                seqid, _, ftype, start, end, _, strand, _, attrs = parts
                attr_dict = dict(
                    item.split("=", 1) for item in attrs.split(";") if "=" in item
                )

                feature = {
                    "seqid": seqid,
                    "type": ftype,
                    "start": int(start),
                    "end": int(end),
                    "strand": strand,
                    "attributes": attr_dict,
                }

                for key in [
                    "ID",
                    "locus_tag",
                    "Name",
                    "gene",
                    "old_locus_tag",
                    "protein_id",
                ]:
                    if val := attr_dict.get(key):
                        for variant in [
                            val,
                            f"gene-{val}",
                            f"cds-{val}",
                            val.replace("_", "~"),
                            val.replace("~", "_"),
                        ]:
                            self._index[variant] = feature

    def get(self, gene_id: str) -> typing.Optional[typing.Dict]:
        """Get feature by gene ID with O(1) lookup.

        Args:
            gene_id: Gene identifier to look up.

        Returns:
            Feature dictionary if found, None otherwise.
        """
        return self._index.get(gene_id)

    def __contains__(self, gene_id: str) -> bool:
        """Check if gene ID exists in index.

        Args:
            gene_id: Gene identifier to check.

        Returns:
            True if gene ID exists, False otherwise.
        """
        return gene_id in self._index


def read_fasta(
    file_path: pathlib.Path, tar_cache: typing.Optional[TarCache] = None
) -> typing.Iterable[typing.Tuple[str, str, str]]:
    """Stream FASTA records from file.

    Handles both plain and gzip-compressed FASTA files.

    Args:
        file_path: Path to FASTA file (.fasta, .fa, .fna, or .gz).
        tar_cache: Optional TarCache instance for caching tar members.

    Yields:
        Tuple of (sequence_id, full_header, sequence_string).
    """
    with smart_open(file_path, tar_cache=tar_cache) as reader:
        with io.TextIOWrapper(reader, encoding="utf-8") as text_reader:
            yield from Bio.SeqIO.parse(text_reader, "fasta")


class ProteinIndex:
    """Lazy-loading protein sequence index for efficient lookup."""

    def __init__(
        self, protein_fasta: pathlib.Path, tar_cache: typing.Optional[TarCache] = None
    ):
        """Initialize protein index without loading sequences.

        Args:
            protein_fasta: Path to protein FASTA file.
            tar_cache: Optional TarCache instance for caching tar members.
        """
        self.path = protein_fasta
        self._tar_cache = tar_cache
        self._sequences: typing.Dict[str, str] = {}
        self._headers: typing.Dict[str, str] = {}
        self._loaded = False

    def _ensure_loaded(self, gene_ids: typing.Optional[set] = None):
        """Load protein sequences on demand.

        Args:
            gene_ids: Optional set of gene IDs to load. If None, loads all.
        """
        if self._loaded:
            return

        for record in read_fasta(self.path, tar_cache=self._tar_cache):
            if gene_ids is None or record.id in gene_ids:
                self._sequences[record.id] = str(record.seq)
                self._headers[record.id] = record.description

        self._loaded = True

    def load_subset(self, gene_ids: set):
        """Pre-load only specific proteins by ID.

        Args:
            gene_ids: Set of gene IDs to load.
        """
        self._ensure_loaded(gene_ids)

    def get(self, protein_id: str) -> typing.Optional[str]:
        """Get protein sequence by ID.

        Args:
            protein_id: Protein identifier.

        Returns:
            Protein sequence if found, None otherwise.
        """
        if not self._loaded:
            self._ensure_loaded()
        return self._sequences.get(protein_id)

    def get_with_fallback(self, gene_id: str) -> typing.Optional[str]:
        """Get protein sequence with fallback to header search.

        Args:
            gene_id: Gene identifier to search for.

        Returns:
            Protein sequence if found, None otherwise.
        """
        seq = self.get(gene_id)
        if seq:
            return seq

        with smart_open(self.path) as reader:
            with io.TextIOWrapper(reader, encoding="utf-8") as text_reader:
                for record in Bio.SeqIO.parse(text_reader, "fasta"):
                    for attr in ["locus_tag", "ID", "Name", "gene"]:
                        if re.search(rf"\[{attr}=({re.escape(gene_id)})\]", record.description):
                            return str(record.seq)

        return None


class GenomeContext:
    """Data container for genome/MAG file paths and metadata."""

    def __init__(
        self,
        genome_id: typing.Optional[str],
        cluster_tsv: pathlib.Path,
        gff_file: pathlib.Path,
        genome_fasta: pathlib.Path,
        protein_fasta: pathlib.Path,
        column_mapping: typing.Dict[str, str],
        system_loader: typing.Callable[[pathlib.Path, Console], pandas.DataFrame],
    ):
        """Initialize genome context.

        Args:
            genome_id: Genome identifier (generates UUID if None).
            cluster_tsv: Path to cluster metadata TSV file.
            gff_file: Path to GFF annotation file.
            genome_fasta: Path to genome FASTA file.
            protein_fasta: Path to protein FASTA file.
            column_mapping: Column name mapping for TSV parsing.
            system_loader: Function to load and filter systems.
        """
        self.genome_id = genome_id if genome_id else str(uuid.uuid4())[:8]
        self.cluster_tsv = pathlib.Path(cluster_tsv)
        self.gff_file = pathlib.Path(gff_file)
        self.genome_fasta = pathlib.Path(genome_fasta)
        self.protein_fasta = pathlib.Path(protein_fasta)
        self.column_mapping = column_mapping
        self.system_loader = system_loader

        self.missing_files = [
            f"{name}: {path}"
            for path, name in [
                (self.cluster_tsv, "cluster_tsv"),
                (self.gff_file, "gff_file"),
                (self.genome_fasta, "genome_fasta"),
                (self.protein_fasta, "protein_fasta"),
            ]
            if not path.exists() and "tar.gz" not in str(path)
        ]
    def __repr__(self):
        """Return string representation."""
        return (
            f"<GenomeContext "
            f"genome_id={self.genome_id!r} "
            f"files={4 - len(self.missing_files)}/4>"
        )

    def is_valid(self) -> bool:
        """Check if all required files exist.

        Returns:
            True if all files are present, False otherwise.
        """
        return len(self.missing_files) == 0


class GenomeResources:
    """Manages lazy-loading and caching of genome resources."""

    def __init__(self, context: GenomeContext, console: Console):
        """Initialize genome resources manager.

        Args:
            context: Genome context with file paths and adapter.
            console: Rich console for logging.
        """
        self.context = context
        self.console = console
        self._cluster_df: typing.Optional[pandas.DataFrame] = None
        self._protein_idx: typing.Optional[ProteinIndex] = None
        self._gff_db: typing.Optional[GFFIndex] = None
        self._coordinates_cache: typing.Optional[typing.List[SystemCoordinates]] = None
        self._tar_cache = TarCache()

    def __enter__(self):
        """Enter context manager."""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit context manager and cleanup resources."""
        self._tar_cache.cleanup()
        return False

    @property
    def cluster_df(self) -> pandas.DataFrame:
        """Load and filter clusters TSV.

        Returns:
            Polars DataFrame with filtered systems.
        """
        if self._cluster_df is not None:
            return self._cluster_df

        df = self.context.system_loader(self.context.cluster_tsv, self.console)

        col_map = self.context.column_mapping
        cluster_col = col_map["cluster_id"]

        dup = df.duplicated(cluster_col)
        n_dup = dup.sum()
        if n_dup > 0:
            dup_ids = df[dup][cluster_col]
            self.console.print(
                f"[bold yellow]{'Warning':>12}[/] {n_dup} duplicate cluster/s in [bold cyan]{self.context.genome_id}[/]: "
                f"[cyan]{', '.join(dup_ids[:5])}{'...' if n_dup > 5 else ''}[/]"
            )
        # if (
        #     n_dup := df.filter(pl.col(cluster_col).is_duplicated())
        #     .select(cluster_col)
        #     .unique()
        #     .height
        # ) > 0:
        #     dup_ids = (
        #         df.filter(pl.col(cluster_col).is_duplicated())
        #         .select(cluster_col)
        #         .unique()
        #         .to_series()
        #         .to_list()
        #     )
        #     self.console.print(
        #         f"[bold yellow]{'Warning':>12}[/] {n_dup} duplicate cluster/s in [bold cyan]{self.context.genome_id}[/]: "
        #         f"[cyan]{', '.join(dup_ids[:5])}{'...' if n_dup > 5 else ''}[/]"
        #     )
        #     df = df.unique(subset=[cluster_col], keep="first")
            df = df[~dup]

        self._cluster_df = df
        return self._cluster_df

    @property
    def protein_idx(self) -> ProteinIndex:
        """Get protein index.

        Returns:
            ProteinIndex instance.
        """
        if self._protein_idx is None:
            self._protein_idx = ProteinIndex(
                self.context.protein_fasta, tar_cache=self._tar_cache
            )
        return self._protein_idx

    @property
    def gff_db(self) -> GFFIndex:
        """Get GFF index.

        Returns:
            GFFIndex instance.
        """
        if self._gff_db is None:
            self._gff_db = GFFIndex(self.context.gff_file, tar_cache=self._tar_cache)
        return self._gff_db

    @property
    def coordinates(self) -> typing.List[SystemCoordinates]:
        """Build and cache system coordinates.

        Returns:
            List of SystemCoordinates for all systems.
        """
        if self._coordinates_cache is None:
            self._coordinates_cache = self._build_coordinates()
        return self._coordinates_cache

    def _build_coordinates(self) -> typing.List[SystemCoordinates]:
        """Parse coordinates from cluster DataFrame.

        Returns:
            List of SystemCoordinates for each cluster.
        """
        coordinates = []
        for row in self.cluster_df.itertuples():
            coord = self._parse_system_coordinates(row)
            coordinates.append(coord)
        return coordinates

    def _parse_system_coordinates(self, row: dict) -> SystemCoordinates:
        """Parse coordinates for a single system.

        Uses adapter's column mapping for format flexibility.

        Args:
            row: Dictionary containing system data from TSV row.

        Returns:
            SystemCoordinates instance.
        """
        col_map = self.context.column_mapping
        cluster_id = getattr(row, col_map['cluster_id'])

        sys_beg_gene = getattr(row, col_map['start_gene'])
        sys_end_gene = getattr(row, col_map['end_gene'])

        if sys_beg_gene is None or sys_end_gene is None:
            return self._invalid_coord(
                cluster_id,
                [],
                f"Missing '{col_map['start_gene']}' or '{col_map['end_gene']}' columns in systems TSV",
            )

        try:
            gene_list = [
                g.strip()
                for g in str(getattr(row, col_map["genes_in_cluster"])).split(",")
                if g.strip()
            ]
        except (KeyError, AttributeError):
            return self._invalid_coord(
                cluster_id,
                [],
                f"Missing '{col_map['genes_in_cluster']}' column",
            )

        if not gene_list:
            return self._invalid_coord(cluster_id, [], "Empty gene list")

        beg_feature = self.gff_db.get(sys_beg_gene)
        end_feature = self.gff_db.get(sys_end_gene)

        if beg_feature is None:
            return self._invalid_coord(
                cluster_id, gene_list, f"Start gene '{sys_beg_gene}' not found in GFF"
            )

        if end_feature is None:
            return self._invalid_coord(
                cluster_id, gene_list, f"End gene '{sys_end_gene}' not found in GFF"
            )

        seq_id_beg = beg_feature["seqid"]
        seq_id_end = end_feature["seqid"]

        if seq_id_beg != seq_id_end:
            return self._invalid_coord(
                cluster_id,
                gene_list,
                f"Start and end genes on different contigs/seqs: {seq_id_beg} vs {seq_id_end} for [bold cyan]{str(self.context.genome_id)}[/]",
            )

        seq_id = seq_id_beg

        start = min(
            beg_feature["start"],
            beg_feature["end"],
            end_feature["start"],
            end_feature["end"],
        )
        end = max(
            beg_feature["start"],
            beg_feature["end"],
            end_feature["start"],
            end_feature["end"],
        )

        strand = beg_feature["strand"]

        region_size = end - start + 1
        if region_size > 1e5:
            self.console.print(
                f"[bold yellow]{'Warning':>12}[/] Cluster [bold cyan]{cluster_id}[/] unusually large: {region_size} bp"
            )
        elif region_size < 50:
            self.console.print(
                f"[bold yellow]{'Warning':>12}[/] Cluster [bold cyan]{cluster_id}[/] unusually small: {region_size} bp"
            )

        return SystemCoordinates(
            cluster_id=cluster_id,
            seq_id=seq_id,
            start_coord=start,
            end_coord=end,
            strand=strand,
            genes=gene_list,
            fasta_file=str(self.context.genome_fasta),
            valid=True,
        )

    def _invalid_coord(
        self, cluster_id: str, genes: typing.List[str], error: str, seq_id: str = ""
    ) -> SystemCoordinates:
        """Create an invalid SystemCoordinates object.

        Args:
            cluster_id: Cluster identifier.
            genes: List of gene identifiers.
            error: Error message describing the issue.
            seq_id: Sequence identifier (default: empty string).

        Returns:
            Invalid SystemCoordinates instance with error message.
        """
        self.console.print(
            f"[bold yellow]{'Warning':>12}[/] {error} for cluster [bold cyan]{cluster_id}[/]"
        )
        return SystemCoordinates(
            cluster_id=cluster_id,
            seq_id=seq_id,
            start_coord=0,
            end_coord=0,
            strand="",
            genes=genes,
            fasta_file=str(self.context.genome_fasta),
            valid=False,
            error_msg=error,
        )

    def extract_genome_sequences(
        self,
        verbose: bool = False,
    ) -> typing.Iterator[Cluster]:
        """Extract nucleotide sequences for gene clusters.

        Streams FASTA file to minimize memory usage.

        Args:
            output: Output sink for writing records.
            verbose: Enable detailed logging.

        Returns:
            List of (cluster_id, sequence_length, fasta_file) tuples.
        """
        coordinates = self.coordinates
        valid_coords = [c for c in coordinates if c.valid]

        if not valid_coords:
            self.console.print(
                f"[bold yellow]{'Warning':>12}[/] No valid clusters to extract for [bold cyan]{self.context.genome_id}[/]"
            )
            return []

        contig_groups: typing.Dict[str, typing.List[SystemCoordinates]] = {}
        for coord in valid_coords:
            contig_groups.setdefault(coord.seq_id, []).append(coord)

        num_contigs = len(contig_groups)

        if verbose:
            self.console.print(
                f"[bold blue]{'Processing':>12}[/] {len(valid_coords)} clusters across {num_contigs} contigs/seqs for [bold cyan]{str(self.context.genome_id)}[/]"
            )

        # results = []

        for record in read_fasta(
            self.context.genome_fasta, tar_cache=self._tar_cache
        ):
            if record.id not in contig_groups:
                continue

            if verbose:
                self.console.print(
                    f"[bold blue]{'Loading':>12}[/] contig [blue]{record.id}[/] ({len(contig_groups[record.id])} clusters) for [bold cyan]{self.context.genome_id}[/]"
                )

            for coord in contig_groups[record.id]:
                subseq = str(record.seq[coord.start_coord - 1 : coord.end_coord])
                yield Cluster(coord.cluster_id, subseq, source=coord.fasta_file)

                if verbose:
                    self.console.print(
                        f"[bold blue]{'Extracted':>12}[/] [cyan]{coord.cluster_id}[/] ({len(subseq)} bp) for [bold cyan]{self.context.genome_id}[/]"
                    )

            del contig_groups[record.id]

            if not contig_groups:
                break

        if contig_groups:
            for seq_id in contig_groups:
                self.console.print(
                    f"[bold red]{'Error':>12}[/] Contig {seq_id} not found in genome [bold cyan]{self.context.genome_id}[/]"
                )

    def extract_proteins_from_coordinates(
        self,
        coordinates: typing.List[SystemCoordinates],
        verbose: bool = False,
    ) -> typing.Iterable[Protein]:
        """Extract protein sequences from gene coordinates.

        Args:
            coordinates: List of cluster coordinates.
            output_file: Output file handle for writing sequences.
            verbose: Enable detailed logging.

        Returns:
            Dictionary mapping protein_id to sequence length.
        """
        valid_coords = [c for c in coordinates if c.valid]

        if not valid_coords:
            self.console.print(
                f"[bold yellow]{'Warning':>12}[/] No valid clusters for [bold cyan]{self.context.genome_id}[/]"
            )
            return

        all_gene_ids = set()
        for coord in valid_coords:
            all_gene_ids.update(coord.genes)

        self.protein_idx.load_subset(all_gene_ids)

        total_genes = len(all_gene_ids)

        if verbose:
            self.console.print(
                f"[bold blue]{'Processing':>12}[/] {total_genes} proteins from {len(valid_coords)} clusters for [bold cyan]{str(self.context.genome_id)}[/]"
            )

        n_extracted = 0
        for coord in valid_coords:
            for gene_id in coord.genes:
                if seq := self.protein_idx.get_with_fallback(gene_id):
                    protein_id = f"{coord.cluster_id}__{gene_id}"
                    yield Protein(
                        protein_id,
                        seq,
                        cluster_id=coord.cluster_id,
                    )
                    n_extracted += 1
                else:
                    if verbose:
                        self.console.print(
                            f"[bold yellow]{'Warning':>12}[/] Protein {gene_id} "
                            f"not found for cluster [bold cyan]{coord.cluster_id}[/]"
                        )

        self.console.print(
            f"[bold blue]{'Extracted':>12}[/] {n_extracted} proteins from "
            f"{len(coordinates)} clusters in [bold cyan]{self.context.genome_id}[/]"
        )


class FastaGFFDataset(BaseDataset):
    """Dataset for extracting sequences from FASTA/GFF files."""

    def __init__(
        self,
        inputs: typing.List[pathlib.Path],
        column_mapping: typing.Optional[typing.Dict[str, str]] = None,
    ) -> None:
        """Initialize the FastaGFFDataset class.

        Args:
            inputs: List of input paths (metadata TSV or individual
                files).
            column_mapping: Custom column mapping. If None, uses
                default generic mapping.
        """
        super().__init__()
        self.inputs = inputs

        self.column_mapping = column_mapping or {
            "cluster_id": "cluster_id",
            "start_gene": "start_gene",
            "end_gene": "end_gene",
            "genes_in_cluster": "genes_in_cluster",
        }

        self.verbose: bool = False
        self.gff_cache_dir: typing.Optional[pathlib.Path] = None
        self._metadata_cache_path: typing.Optional[pathlib.Path] = None
        self.cluster_metadata = inputs[0]

    def _create_genome_context(self, row: typing.Dict, genome_id: str) -> GenomeContext:
        """Create GenomeContext from metadata row.

        Args:
            row: Dictionary with file paths from metadata TSV.
            genome_id: Genome identifier.

        Returns:
            GenomeContext instance.
        """
        return GenomeContext(
            genome_id=genome_id,
            cluster_tsv=pathlib.Path(row.cluster_tsv),
            gff_file=pathlib.Path(row.gff_file),
            genome_fasta=pathlib.Path(row.genome_fasta_file),
            protein_fasta=pathlib.Path(row.protein_fasta_file),
            column_mapping=self.column_mapping,
            system_loader=self._load_and_filter_systems,
        )

    def _load_and_filter_systems(
        self, tsv_path: pathlib.Path, console: Console
    ) -> pandas.DataFrame:
        """Load systems TSV file.

        Args:
            tsv_path: Path to systems TSV.
            console: Rich console for logging.

        Returns:
            Polars DataFrame with system data.
        """
        df = pandas.read_csv(tsv_path, sep="\t")
        return df

    def extract_clusters(
        self,
        progress: rich.progress.Progress,
    ) -> typing.Iterable[Cluster]:
        progress.console.print(
            f"[bold blue]{'Using':>12}[/] cluster metadata file: [magenta]{self.cluster_metadata}[/]"
        )

        df = pandas.read_csv(self.cluster_metadata, sep="\t")

        results = []

        genome_count = 0
        task = progress.add_task(
            f"[bold blue]{'Processing':>9}[/] gene clusters", total=len(df)
        )

        for row in df.itertuples():
            genome_id = getattr(row, "genome_id", f"genome_{genome_count:07}")
            genome_count += 1
            progress.update(
                task,
                description=f"[bold blue]{'Processing':>9}[/] genome: [bold cyan]{genome_id}",
            )

            context = self._create_genome_context(row, genome_id)

            if not context.is_valid():
                progress.console.print(
                    f"[bold yellow]{'Missing':>12}[/] files for [bold cyan]{genome_id}[/]"
                )
                progress.update(task, advance=1)
                continue

            with GenomeResources(context, progress.console) as resources:
                coordinates = resources.coordinates
                valid_count = sum(1 for c in coordinates if c.valid)

                progress.update(
                    task,
                    description=f"[bold blue]{'Processing':>9}[/] genome: [bold cyan]{genome_id}[/] - validated {valid_count}/{len(coordinates)} clusters",
                )

                n_extracted = 0
                for cluster in resources.extract_genome_sequences(self.verbose):
                    yield cluster
                    n_extracted += 1

                contig_groups = collections.defaultdict(list)
                for coord in coordinates:
                    if coord.valid:
                        contig_groups[coord.seq_id].append(coord)
                num_contigs = len(contig_groups)

                progress.console.print(
                    f"[bold blue]{'Extracted':>12}[/] {n_extracted} "
                    f"gene clusters across {num_contigs} contigs/seqs for "
                    f"[bold cyan]{context.genome_id}[/]"
                )

                for coord in coordinates:
                    if coord.valid:
                        results.append(
                            (
                                coord.cluster_id,
                                coord.end_coord - coord.start_coord + 1,
                                coord.fasta_file,
                            )
                        )

            progress.update(task, advance=1)

        progress.remove_task(task)
        progress.console.print(
            f"[bold green]{'Extracted':>12}[/] {len(results)} gene clusters "
            f"from {len(df)} genomes/MAGs"
        )

    def extract_proteins(
        self,
        progress: rich.progress.Progress,
        cluster_ids: typing.Container[str],
    ) -> typing.Dict[str, int]:

        df = pandas.read_csv(self.cluster_metadata, sep="\t")

        genome_count = 0
        task = progress.add_task(
            f"[bold blue]{'Processing':>9}[/] protein sequences", total=len(df)
        )

        for row in df.itertuples():
            genome_id = getattr(row, "genome_id", f"genome_{genome_count:07}")
            genome_count += 1
            progress.update(
                task,
                description=f"[bold blue]{'Processing':>9}[/] genome: [bold cyan]{genome_id}",
            )

            context = self._create_genome_context(row, genome_id)

            if not context.is_valid():
                progress.console.print(
                    f"[bold yellow]{'Missing':>12}[/] files for {genome_id}"
                )
                progress.update(task, advance=1)
                continue

            with GenomeResources(context, progress.console) as resources:
                coordinates = resources.coordinates
                if cluster_ids:
                    coordinates = [
                        c
                        for c in coordinates
                        if c.cluster_id in cluster_ids
                    ]
                yield from resources.extract_proteins_from_coordinates(
                    coordinates,
                    verbose=self.verbose,
                )

            progress.update(task, advance=1)

        progress.remove_task(task)

    def _log_protein_summary(
        self,
        progress: rich.progress.Progress,
        results: typing.Dict,
        representatives: typing.Optional[typing.Container[str]],
    ):
        """Log summary of protein extraction results.

        Args:
            progress: Rich progress bar instance.
            results: Dictionary of extracted proteins.
            representatives: Container of representative cluster IDs.
        """
        rep_count = "all"
        if representatives:
            if hasattr(representatives, "__len__"):
                rep_count = str(len(representatives))
            else:
                rep_count = "specified"

        progress.console.print(
            f"[bold green]{'Extracted':>12}[/] {len(results):,} proteins from {rep_count} representative gene clusters"
        )



import abc
import typing
from typing import Literal

import numpy
import scipy.sparse
from scipy.cluster.hierarchy import fcluster, DisjointSet

from .hca import linkage, manhattan, manhattan_pair


class ClusteringStrategy(abc.ABC):
    """An abstract clustering strategy to cluster compositional data.
    """
    method: str

    @abc.abstractmethod
    def cluster(
        self,
        X: scipy.sparse.csr_matrix,
        weights: typing.Optional[numpy.ndarray] = None,
    ) -> numpy.ndarray:
        r"""Cluster the given observations.

        Arguments:
            X (`scipy.sparse.csr_matrix`): A matrix of shape :math:`m \times n`
                with compositional data.
            weights (`numpy.ndarray` or `None`): The weights (of shape
                :math:`n`) to use for computing distances. If `None`,
                use uniform weights.

        Returns:
            `numpy.ndarray`: A flat array of shape :math:`m` which assigns
            an arbitrary cluster number to each observation of ``X``
            (see `scipy.cluster.hierarchy.fcluster` documentation).

        """


class HierarchicalClustering(ClusteringStrategy):
    """A clustering strategy implementing hierarchical clustering.
    """

    def __init__(
        self,
        *,
        method: Literal["average", "single", "complete", "weighted", "centroid", "median", "ward"],
        distance: float = 0.8,
        precision: Literal["half", "single", "double"] = "double",
        jobs: int = 1,
    ):
        """Create a new hierarchical clustering strategy.

        Arguments:
            method (`str`): The name of the linkage method to use: *average*,
                *single*, *complete*, *weighted*, *centroid*, *median* or
                *ward*.
            distance (`float`): The distance cutoff for the flat clusters
                created with `scipy.cluster.hierarchy.fcluster`.
            precision (`str`): The floating-point precision to use: *half*,
                *single* or *double*. Note that changing precision (in
                particular with half-precision) may lead to differences in
                the generated clustering.
            jobs (`int`): The number of parallel threads to use to perform
                the pairwise distance computation.

        """
        self.method = method
        self.distance = distance
        self.precision = precision
        self.jobs = jobs

    def _compute_distances(
        self,
        X: scipy.sparse.csr_matrix,
        weights: typing.Optional[numpy.ndarray] = None,
    ) -> numpy.ndarray:
        # check matrix format
        if not isinstance(X, scipy.sparse.csr_matrix):
            raise TypeError(f"expected csr_matrix, got {type(X).__name__}")

        # make sure the sparse matrix has sorted indices (necessary for
        # the distance algorithm to work efficiently)
        if not X.has_sorted_indices:
            X.sort_indices()

        # compute the total length of each observation
        # (used to rescale the absolute distances to a relative range)
        r = X.shape[0]
        if weights is None:
            total = X.sum(axis=1, dtype=numpy.int32).A1
        else:
            total = X @ weights

        # compute manhattan distance on sparse matrix
        distance_vector = numpy.zeros(r * (r - 1) // 2, dtype=self.precision)
        manhattan(
            X.data,
            X.indices,
            X.indptr,
            weights,
            distance_vector,
            threads=self.jobs,
        )
        # ponderate by sum of amino-acid distance
        n = 0
        for i in range(r - 1):
            l = r - (i + 1)
            maxdist = numpy.clip(total[i + 1 :] + total[i], 1, numpy.inf)
            distance_vector[n : n + l] /= maxdist
            n += l
        # enforce distances to be in [0, 1] (slight possibility of >1 due
        # to handling of alignment weights but we can round off)
        return numpy.clip(distance_vector, 0.0, 1.0, out=distance_vector)

    def cluster(
        self,
        X: scipy.sparse.spmatrix,
        weights: typing.Optional[numpy.ndarray] = None,
    ) -> numpy.ndarray:
        if weights is not None and X.shape[1] != weights.shape[0]:
            raise ValueError("inconsistent shapes between X and weights")
        pdist = self._compute_distances(X, weights)
        Z = linkage(pdist, method=self.method)
        return fcluster(Z, criterion="distance", t=self.distance)


class LinearClustering(ClusteringStrategy):
    """A clustering strategy similar to MMseqs2 linear clustering.
    """

    def __init__(
        self,
        *,
        distance: float = 0.8,
    ):
        """Create a new linear clustering strategy.

        Arguments:
            distance (`float`): The distance cutoff to use for clustering
                observations together.

        """
        self.distance = distance
        self.method = "linclust"

    def cluster(
        self,
        X: scipy.sparse.csr_matrix,
        weights: typing.Optional[numpy.ndarray] = None,
    ) -> numpy.ndarray:
        # check matrix format
        if not isinstance(X, scipy.sparse.csr_matrix):
            raise TypeError(f"expected csr_matrix, got {type(X).__name__}")
        if weights is not None and X.shape[1] != weights.shape[0]:
            raise ValueError("inconsistent shapes between X and weights")

        # make sure the sparse matrix has sorted indices (necessary for
        # the distance algorithm to work without returning bogus values)
        if not X.has_sorted_indices:
            X.sort_indices()

        # also compute a csc_matrix form the input to support fast
        # selection of columns in the loop
        X_csc = X.tocsc()

        # compute the total length of each observation
        # (used to rescale the absolute distances to a relative [0, 1] range)
        r = X.shape[0]
        if weights is None:
            total = X.sum(axis=1, dtype=numpy.int32).A1
        else:
            total = X @ weights

        # use scipy DisjointSet / UnionFind to record clustering
        ds = DisjointSet(range(r))

        # extract columns with more than one row so we only iterate on those
        n_obs = (X_csc != 0).sum(axis=0)
        cols = numpy.where(n_obs > 1)[1]

        # iterate on subsets formed by protein membership and cluster together
        # all observations with relative weighted Manhattan under the distance
        # threshold, taking the largest observation (in aa.) as the centroid
        for y in cols:
            # extract which gene clusters contain protein 'y'
            # (indexing is faster with the compressed-sparse-column matrix)
            mask = X_csc[:, y]
            if mask.nnz <= 1:
                continue
            # extract the row subset indices directly from the CSC (r x 1) array
            indices = mask.indices
            # find t he largest gene cluster of the row subset
            centroid = indices[numpy.argmax(total[indices])]
            # compare every other gene clusters to the reference
            # TODO: implement a Rust function to for manhattan distance
            #       a la cdist so we can do distance of centroid to every
            #       other query efficiently in a single call?
            for query in indices:
                d = manhattan_pair(X.data, X.indices, X.indptr, weights, query, centroid)
                d /= numpy.clip(total[query] + total[centroid], 1.0, numpy.inf)
                if d <= self.distance:
                    ds.merge(query, centroid)

        # extract the cluster indices from the disjoint set
        # NOTE: implementation not super efficient at the moment as
        #       `ds.subsets` returns a list of set that we need to convert
        #       again, maybe prohibitive for very large datasets
        flat = numpy.zeros(r, dtype=numpy.int32)
        for i, subset in enumerate(ds.subsets(), 1):
            flat[list(subset)] = i

        return flat


def default_strategy() -> ClusteringStrategy:
    return HierarchicalClustering(
        method="average",
        distance=0.8,
        precision="double",
    )
